"""Functions that calculate curve data for cross validation plots."""
from typing import List
import warnings

import numpy as np
import pandas as pd
from sklearn import metrics


def calculate_roc_curve(curve_data):
    """Calculates ROC curve on the folds

    Args:
        curve_data (dict) : dict of curves produced by
            BaseEvaluationPlotDataExtractor.calculate_curve_data
    Returns:
        dict[str, list[np.ndarray]]: Keys being "FPR", "TPR" and "AUC" (ROC metrics)
            and values are a list the size of number of folds with the evaluation of each fold.
    """

    for curve_name in curve_data.keys():
        curve_data[curve_name]["FPR"] = curve_data[curve_name].pop("first_ret_value")
        curve_data[curve_name]["TPR"] = curve_data[curve_name].pop("second_ret_value")
        curve_data[curve_name]["AUC"] = curve_data[curve_name].pop("area")
    return curve_data


def calculate_pr_curve(curve_data, targets):
    """Calculates precision-recall curve on the folds.

    Args:
        curve_data (dict) : dict of curves produced by
            BaseEvaluationPlotDataExtractor.calculate_curve_data
        targets (pd.Series): True labels.

    Returns:
        dict[str, list[np.ndarray]]: Keys being "Precision", "Recall" and "AP" (PR metrics)
            and values are a list the size of number of folds with the
            evaluation of each fold.
            Additional "prevalence" key, with positive-label "prevalence" is added
            to be used by the chance curve.
    """

    for curve_name in curve_data.keys():
        curve_data[curve_name]["Precision"] = curve_data[curve_name].pop(
            "first_ret_value"
        )
        curve_data[curve_name]["Recall"] = curve_data[curve_name].pop(
            "second_ret_value"
        )
        curve_data[curve_name]["AP"] = curve_data[curve_name].pop("area")
    curve_data["prevalence"] = targets.value_counts(normalize=True).loc[targets.max()]
    return curve_data


def calculate_performance_curve_data_on_folds(
    folds_predictions,
    folds_targets,
    sample_weights=None,
    area_metric=metrics.roc_auc_score,
    curve_metric=metrics.roc_curve,
    pos_label=None,
):
    """Calculates performance curves of the predictions across folds.

    Args:
        folds_predictions (list[pd.Series]): Score prediction (as in continuous output
            of classifier, `predict_proba` or `decision_function`) for every fold.
        folds_targets (list[pd.Series]): True labels for every fold.
        sample_weights (list[pd.Series] | None): weight for each sample for every fold.
        area_metric (callable): Performance metric of the area under the curve.
        curve_metric (callable): Performance metric returning 3 output vectors - metric1, metric2
            and thresholds.
            Where metric1 and metric2 depict the curve when plotted on x-axis and y-axis.
        pos_label: What label in `targets` is considered the positive label.

    Returns:
        (list[np.ndarray], list[np.ndarray], list[np.ndarray], list[float]):
            For every fold, the calculated metric1 and metric2 (the curves), the thresholds and the
            area calculations.
    """
    sample_weights = (
        [None] * len(folds_predictions) if sample_weights is None else sample_weights
    )
    # Scikit-learn precision_recall_curve and roc_curve do not return values in a consistent way.
    # Namely, roc_curve returns `fpr`, `tpr`, which correspond to x_axis, y_axis,
    # whereas precision_recall_curve returns `precision`, `recall`,
    # which correspond to y_axis, x_axis.
    # That's why this function will return the values the same order as Scikit's curves,
    # leaving it up to the caller to put labels on what those return values actually are
    # (specifically, whether they're x_axis or y-axis)
    first_ret_folds, second_ret_folds, threshold_folds, area_folds = [], [], [], []
    for fold_prediction, fold_target, fold_weights in zip(
        folds_predictions, folds_targets, sample_weights
    ):
        first_ret_fold, second_ret_fold, threshold_fold = curve_metric(
            fold_target,
            fold_prediction,
            pos_label=pos_label,
            sample_weight=fold_weights,
        )
        try:
            area_fold = area_metric(
                fold_target, fold_prediction, sample_weight=fold_weights
            )
        except ValueError as v:  # AUC cannot be evaluated if targets are constant
            warnings.warn(f"metric {area_metric.__name__} could not be evaluated")
            warnings.warn(str(v))
            area_fold = np.nan

        first_ret_folds.append(first_ret_fold)
        second_ret_folds.append(second_ret_fold)
        threshold_folds.append(threshold_fold)
        area_folds.append(area_fold)
    return area_folds, first_ret_folds, second_ret_folds, threshold_folds


def calculate_curve_data_binary_outcome(
    folds_predictions,
    targets,
    curve_metric,
    area_metric,
    stratify_by=None,
):
    """Calculate different performance (ROC or PR) curves

    Args:
        folds_predictions (list[pd.Series]): Predictions for each fold.
        targets (pd.Series): True labels
        curve_metric (callable): Performance metric returning 3 output vectors - metric1,
            metric2 and thresholds. Where metric1 and metric2 depict the curve
            when plotted on x-axis and y-axis.
        area_metric (callable): Performance metric of the area under the curve.
        stratify_by (pd.Series): Group assignment to stratify by.

    Returns:
        dict[str, dict[str, list[np.ndarray]]]: Evaluation of the metric
            for each fold and for each curve.
            One curve for each group level in `stratify_by`.
            On general: {curve_name: {metric1: [evaluation_fold_1, ...]}}.
            For example: {"Treatment=1": {"FPR": [FPR_fold_1, FPR_fold_2, FPR_fold_3]}}
    """
    # folds_targets = [targets.loc[p.index] for p in folds_predictions]
    # folds_stratify_by = [stratify_by.loc[p.index] for p in folds_predictions]

    stratify_values = sorted(set(stratify_by))
    curve_data = {}
    for stratum_level in stratify_values:
        # Slice data for that stratum level across the folds:
        folds_stratum_predictions, folds_stratum_targets = [], []
        for fold_predictions in folds_predictions:
            # Extract fold:
            fold_targets = targets.loc[fold_predictions.index]
            fold_stratify_by = stratify_by.loc[fold_predictions.index]
            # Extract stratum:
            mask = fold_stratify_by == stratum_level
            fold_predictions = fold_predictions.loc[mask]
            fold_targets = fold_targets.loc[mask]
            # Save:
            folds_stratum_predictions.append(fold_predictions)
            folds_stratum_targets.append(fold_targets)

        (
            area_folds,
            first_ret_folds,
            second_ret_folds,
            threshold_folds,
        ) = calculate_performance_curve_data_on_folds(
            folds_stratum_predictions,
            folds_stratum_targets,
            None,
            area_metric,
            curve_metric,
        )

        curve_data[f"Treatment={stratum_level}"] = {
            "first_ret_value": first_ret_folds,
            "second_ret_value": second_ret_folds,
            "Thresholds": threshold_folds,
            "area": area_folds,
        }
    return curve_data


def calculate_curve_data_propensity(
    fold_predictions: List[
        "causallib.evaluation.weight_predictor.PropensityPredictions"
    ],
    targets,
    curve_metric,
    area_metric,
):
    """Calculate different performance (ROC or PR) curves

    Args:
        fold_predictions (list[PropensityEvaluatorPredictions]):
            Predictions for each fold.
        targets (pd.Series): True labels
        curve_metric (callable): Performance metric returning 3 output vectors - metric1,
            metric2 and thresholds. Where metric1 and metric2 depict the curve when plotted
            on x-axis and y-axis.
        area_metric (callable): Performance metric of the area under the curve.
        **kwargs:

    Returns:
        dict[str, dict[str, list[np.ndarray]]]: Evaluation of the metric
            for each fold and for each curve.
            3 curves:
                * "unweighted" (regular)
                * "weighted" (weighted by inverse propensity)
                * "expected" (duplicated population, weighted by propensity)
            On general: {curve_name: {metric1: [evaluation_fold_1, ...]}}.
            For example: {"weighted": {"FPR": [FPR_fold_1, FPR_fold_2, FPR_fold3]}}
    """

    curves_sample_weights = {
        "unweighted": [None for _ in fold_predictions],
        "weighted": [
            fold_predictions.weight_by_treatment_assignment
            for fold_predictions in fold_predictions
        ],
        "expected": [
            pd.concat([fold_predictions.propensity, 1 - fold_predictions.propensity])
            for fold_predictions in fold_predictions
        ],
    }
    curves_folds_targets = [
        targets.loc[fold_predictions.weight_by_treatment_assignment.index]
        for fold_predictions in fold_predictions
    ]
    curves_folds_targets = {
        "unweighted": curves_folds_targets,
        "weighted": curves_folds_targets,
        "expected": [
            pd.concat([
                pd.Series(data=targets.max(), index=fold_predictions.propensity.index),
                pd.Series(data=targets.min(), index=fold_predictions.propensity.index)
            ])
            for fold_predictions in fold_predictions
        ],
    }
    fold_predictions = {
        "unweighted": [
            fold_predictions.propensity for fold_predictions in fold_predictions
        ],
        "weighted": [
            fold_predictions.propensity for fold_predictions in fold_predictions
        ],
        "expected": [
            pd.concat([fold_predictions.propensity, fold_predictions.propensity])
            for fold_predictions in fold_predictions
        ],
    }
    # Expected curve duplicates the population, basically concatenating so that:
    # prediction = [p, p], target = [1, 0], weights = [p, 1-p]

    curve_data = {}
    for curve_name in curves_sample_weights:
        sample_weights = curves_sample_weights[curve_name]
        folds_targets = curves_folds_targets[curve_name]
        folds_predictions = fold_predictions[curve_name]

        (
            area_folds,
            first_ret_folds,
            second_ret_folds,
            threshold_folds,
        ) = calculate_performance_curve_data_on_folds(
            folds_predictions,
            folds_targets,
            sample_weights,
            area_metric,
            curve_metric,
        )

        curve_data[curve_name] = {
            "first_ret_value": first_ret_folds,
            "second_ret_value": second_ret_folds,
            "Thresholds": threshold_folds,
            "area": area_folds,
        }

    # Rename keys (as will be presented as curve labels in legend)
    curve_data["Propensity"] = curve_data.pop("unweighted")
    curve_data["Weighted"] = curve_data.pop("weighted")
    curve_data["Expected"] = curve_data.pop("expected")
    return curve_data



from sklearn.model_selection import GridSearchCV as skGridSearchCV
from sklearn.model_selection import RandomizedSearchCV as skRandomizedSearchCV

from .search import causalize_searcher
from .split import TreatmentOutcomeStratifiedKFold
from .split import TreatmentStratifiedKFold

GridSearchCV = causalize_searcher(skGridSearchCV)
RandomizedSearchCV = causalize_searcher(skRandomizedSearchCV)



# coding: utf-8

# (C) Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Created on Sept 25, 2019

import torch
import torch.nn as nn
import numpy as np

from collections import Counter
from torch import logsumexp as logsumexp  # pylint: disable=E0611
from sklearn.metrics import roc_auc_score, average_precision_score

from .hemm_metrics import wass, mmd2_lin, mmd2_rbf
from copy import deepcopy


class HEMM(torch.nn.Module):
    """This is the model defintion. The model has two parts:

      1. the subgroup discovery component 
      2. the outcome prediction from the subgroup assignment and the interaction with confounders through an MLP. 
    
    Args:
        D_in: the size of the features of the data
        K: number of components to discover.
        homo (bool): Flag to specify if the final outcome model is same for each discovered subgroup.
                    Default is True ie. same outcome model is used for each subgroup.          
        mu: initialize the components with means of the training data.
        std: initialize the components with std dev of the training data.
        bc: the first bc components are considered bernoulli variables
        lamb: strength of the beta(0.5, 0.5) prior on the bernoulli variables
        spread: how far should the components be initialized from there means.
        outcomeModel: 'linear' to specify a linear outcome function. Or pass another Torch.model as the outcome model.
        sep_heads: Setting false will force the adjustment of Confounding to be same independent of treatment assignment.
    """

    def __init__(self, D_in, K, homo=True, mu=None, std=None, bc=0, lamb=0.0001, spread=0.1, outcomeModel='linear', sep_heads=True):

        super(HEMM, self).__init__()
        self.bc = bc
        self.lamb = lamb
        self.homo = homo
        self.K = K 
        self.sep_heads = sep_heads
        
        self.tc = None
        
        lindim = D_in

        if outcomeModel == 'linear':
            outcomeModel = nn.Linear(lindim, 2)

        if mu is not None:
            p = torch.from_numpy(np.repeat(mu[:, bc:], self.K, axis=0))
            mu = torch.from_numpy(np.repeat(mu[:, :bc], self.K, axis=0))
        else:
            p = 0.5 * torch.ones(self.K, D_in - bc)
            mu = torch.zeros(self.K, bc)

        if std is not None:
            std = torch.from_numpy(np.repeat(std[:, :bc], self.K, axis=0))
        else:
            std = torch.ones(self.K, bc)

        for i in range(self.K):
            mu[i] = mu[i] + spread * (torch.rand_like(mu[i]) - 0.5)
            std[i] = std[i] + spread * (torch.rand_like(std[i]) - 0.5)
            p[i] = p[i] + spread * (torch.rand_like(p[i]) - 0.5)

        self.mu = nn.ParameterList(nn.Parameter(mu[i]) for i in range(self.K))
        self.std = nn.ParameterList(nn.Parameter(std[i]) for i in range(self.K))
        self.p = nn.ParameterList(nn.Parameter(p[i]) for i in range(self.K))
        self.alph = nn.Parameter(torch.ones(self.K))

        treat = torch.abs(torch.rand(self.K))
        self.treat = nn.Parameter(treat.double())
        if homo:
            expert = [deepcopy(outcomeModel) for i in range(1)]
        else:
            expert = [deepcopy(outcomeModel) for i in range(self.K)]

        self.expert = nn.ModuleList(expert)

    @staticmethod
    def gaussian_pdf(x, mu, std):
        mu = mu.unsqueeze(0)
        std = std.unsqueeze(0)

        gauss_ = -torch.log(std) - torch.div((x - mu.expand(x.shape)) ** 2, 2 * (std ** 2))
        gauss_ = torch.sum(gauss_, dim=1) - 0.9189 * x.shape[1]  # TODO: what's with the 0.9189?

        return gauss_

    @staticmethod
    def bernoulli_pdf(x, mu):
        loss = nn.BCELoss(reduction='none')

        mu = mu.unsqueeze(0)
        mu = torch.clamp(mu, min=1e-3, max=1 - 1e-3)

        bern_ = -loss(mu.expand(x.shape), x)

        return torch.sum(bern_, dim=1)

    def regularization(self, alpha, beta):
        lamb = self.lamb
        p = []
        for i in range(self.K):
            p.append(torch.clamp(self.p[i], min=1e-3, max=1 - 1e-3))

        p = torch.stack(p, 1)

        cost = ((alpha - 1) * torch.log(p)) + ((beta - 1) * torch.log(1 - p))
        cost = cost.sum()

        return -lamb * cost
    
    def forward(self, x, t, soft=True, infer=True, response='bin'):
        
        if self.sep_heads:
            selector = t.cpu().data.numpy().astype('int').tolist()
        else:
            selector = torch.zeros_like(t).cpu().data.numpy().astype('int').tolist()
            
        selector2 = range(len(selector))

        # TODO: Consider a slightly more general binary-variable mask vector indicated which columns are binary,
        #       rather than force certain structure on the input.
        xg, xb = x[:, :self.bc], x[:, self.bc:]

        # compute q(Z)
        gate_output = []
        for i in range(self.K):
            z_i = self.gaussian_pdf(xg, self.mu[i], self.std[i])
            z_i += self.bernoulli_pdf(xb, self.p[i])
            gate_output.append(z_i)

        gate_output = torch.stack(gate_output, 1)
        gate_output = gate_output + self.alph.expand(gate_output.shape)
        gate_output_ = gate_output
        gate_sum = logsumexp(gate_output, dim=1, keepdim=True)

        lgate_output = gate_output - gate_sum
        gate_output = torch.exp(lgate_output)

        rs = x.shape[0]
        cs = self.treat.shape[0]  # size of K  # TODO: why not use self.K instead?
        treat = self.treat.expand((rs, cs))

        # TODO: Consider renaming `infer` to a more informative variable name.

        expert_output = []
        for i in range(self.K):
            # calculate teffect (main effect due to treatment and coefficient) and add to output of outcomeModel
            teffect = torch.mul(t, treat[:, i])
            if self.homo:
                cur_expert = self.expert[0]
            else:
                cur_expert = self.expert[i]
            cur_expert_output = cur_expert(x)[selector2, selector] + teffect
            if infer:
                if response == 'bin':
                    cur_expert_output = nn.Sigmoid()(cur_expert_output)

            expert_output.append(cur_expert_output)
        expert_output = torch.stack(expert_output, 1)
        
        if not infer:
            return gate_output_, lgate_output, expert_output
        
        else:
            if soft:
                # TODO: consider vectorized math below: `gate_output.mul(expert_output).sum(dim=1)`
                output = torch.zeros_like(x[:, 0])
                for i in range(self.K):
                    output += torch.mul(gate_output[:, i], expert_output[:, i])
                return output
            else:
                ridx = torch.tensor(range(x.shape[0]))
                cidx = torch.argmax(gate_output, 1)
                output = expert_output[ridx, cidx]
                return output

    def fit(self, train, epochs=100, batch_size=10, lr=1e-3, wd=0, ltype='log', dev=None, metric='AP', response='bin',
            use_p_correction=True, imb_fun='mmd2_lin', p_alpha=1e-4):
        """This method uses ELBO to perform parameter updates using a first order optimizer routing.

        Args:
            train: (x_, t_, Y_):= Tuple o torch tensors of the input features, treatment and outcome.
            epochs: Max number of epochs.
            batch_size: Batch size for optimizer.
            lr: Learning rate for the optimizer.
            wd (float): Weight decay
            ltype: 'log' for the ELBO.
            dev: Tuple of validation dataset i.e.: torch tensors (x_, t_, Y_).
                If provided, early-stopping criteria using `metric` will be applied.
            metric (str): 'AP' mean average precision, 'AuROC': area under roc curve, 'MSE':mean squared error.
                          If not specified it uses the optimizer cost as the metric.
                          The specified this metric is computed for the training set (or dev set if it is specified)
                          and used to perform early stopping to prevent overfitting if dev set is provided.
            response: Specify if y is continuous ('cont') or binary ('bin').
            use_p_correction (bool): Whether to use population size p(treated) in imbalance penalty (IPM).
            imb_fun (str): Which imbalance penalty to use ('mmd_lin', 'wass').
            p_alpha (float): Imbalance regularization parameter.
        """
        # TODO: Consider rename `dev` to `val` to correspond with Keras `validation_data` parameters `x_val, y_val`.
        #       https://keras.io/models/model/#fit
        x, t, y = train

        if dev is not None:
            xdev, tdev, ydev = dev
            devcost = -float('inf')
        else:
            xdev, tdev, ydev = x, t, y

        optparam = filter(lambda p: p.requires_grad, self.parameters())
        optimizer = torch.optim.Adam(optparam, lr=lr, weight_decay=wd)

        losses = []
        greaters = 0

        for epoch in range(epochs):
            tcost = 0
            for step in range(int(x.shape[0] / batch_size)):
                x_ = x[step * batch_size:(step + 1) * batch_size]
                y_, t_ = y[step * batch_size:(step + 1) * batch_size], t[step * batch_size:(step + 1) * batch_size]
                optimizer.zero_grad()
                cost = self.lcost(x_, t_, y_, ltype=ltype, response=response, use_p_correction=use_p_correction,
                                  imb_fun=imb_fun, p_alpha=p_alpha)
                tcost += cost.cpu().data.numpy()
                cost.backward()
                optimizer.step()

            losses.append(float(tcost) / x.shape[0])

            if dev is not None:
                output = self.forward(xdev, tdev, response=response)
            else:
                output = self.forward(x, t, response=response)

            if response == 'bin':
                if metric == 'AuROC':
                    newdevcost = roc_auc_score(ydev.cpu().data.numpy(), output.cpu().data.numpy())
                elif metric == 'AP':
                    newdevcost = average_precision_score(ydev.cpu().data.numpy(), output.cpu().data.numpy())
                else:
                    newdevcost = -self.lcost(xdev, tdev, ydev, ltype=ltype, response=response,
                                             use_p_correction=use_p_correction, imb_fun=imb_fun, p_alpha=p_alpha)
            elif response == 'cont':
                if metric == 'MSE':
                    newdevcost = -((ydev.cpu().data.numpy() - output.cpu().data.numpy()) ** 2).sum()
                else:
                    newdevcost = -self.lcost(xdev, tdev, ydev, ltype=ltype, response=response,
                                             use_p_correction=use_p_correction, imb_fun=imb_fun, p_alpha=p_alpha)

            if dev is not None:
                if newdevcost < devcost:
                    greaters += 1

                if greaters > 3:
                    break

            devcost = newdevcost
        return losses

    def lcost(self, x_, t_, y_, ltype="log", response='bin', use_p_correction=True, imb_fun=None, p_alpha=1e-4,
              rbf_sigma=0.1, wass_its=20, wass_lambda=10.0):
        """Implements ELBO as the objective function (eq 12 in paper).

        Args:
            x_ (torch.Tensor): Covariate matrix of size (num_subjects, num_features).
            t_ (torch.Tensor): Treatment assignment of size (num_subjects,).
            y_ (torch.Tensor): Outcome of size (num_subjects,).
            ltype: 'log' specifies ELBO
            response: specifies whether outcome `y` is binary ('bin') or continuous ('cont').
            use_p_correction (bool): whether to use population size p(treated) in imbalance penalty (IPM)
            imb_fun (str): which imbalance penalty to use ('mmd_lin', 'mmd_rbf', 'wass', 'wass2')
            p_alpha (float): imbalance regularization parameter
            rbf_sigma (float): RBF MMD sigma
            wass_its (int): Number of iterations in Wasserstein computation.
            wass_lambda (float): Wasserstein lambda.
        """
        if response == 'bin':
            loss = nn.BCEWithLogitsLoss(reduction='none')
        elif response == 'cont':
            loss = nn.MSELoss(reduction='none')
        else:
            raise ValueError("supported response values are 'bin' and 'cont', got '{}' instead".format(response))

        if ltype == 'log':
            # COMPUTE ELBO
            lgate_out_, lgate_out, exp_out = self.forward(x_, t_, infer=False)
            cost = torch.zeros_like(x_[:, 0])

            lexp_out = []
            for i in range(self.K):
                lexp_out.append(-loss(exp_out[:, i], y_))

            lexp_out = torch.stack(lexp_out, 1)
            gate_out = torch.exp(lgate_out)

            cost = torch.sum(torch.mul(gate_out, lexp_out), dim=1)  # eq 12 in paper
            cost = -torch.sum(cost) + self.regularization(0.5, 0.5)

            # IMBALANCE PENALTY (IPM)
            # compute treatment probability
            if use_p_correction:
                p_t = torch.mean(t_)
            else:
                p_t = 0.5

            if imb_fun == 'wass':
                imb_dist, imb_mat = wass(x_, t_, p_t, sq=False, its=wass_its, lam=wass_lambda)
                imb_error = p_alpha * imb_dist
            elif imb_fun == 'wass2':
                imb_dist, imb_mat = wass(x_, t_, p_t, sq=True, its=wass_its, lam=wass_lambda)
                imb_error = p_alpha * imb_dist
            elif imb_fun == 'mmd2_lin':
                imb_dist = mmd2_lin(x_, t_, p_t)
                imb_error = p_alpha * imb_dist
            elif imb_fun == 'mmd2_rbf':
                imb_dist = mmd2_rbf(x_, t_, p_t, rbf_sigma)
                imb_error = p_alpha * imb_dist
            elif imb_fun is None:
                imb_error = 0
            else:
                raise ValueError("supported imb_fun values are ['wass', 'wass2', 'mmd2_lin', 'mmd2_rbf'], "
                                 "got '{}' instead".format(imb_fun))

            cost += imb_error
        else:
            output = self.forward(x_, t_)
            cost = loss(output, y_)
            cost = cost.sum()

        return cost

    def group_sizes(self, X):
        """Returns the number of data points assigned to each subgroup.
        
        Args:
            X (torch.Tensor): Covariate matrix of size (num_subjects, num_features).

        Returns: 
            collections.Counter: giving size for each group
        """
        group = self.get_groups(X)
        counter = Counter(group)
        return counter

    def estimate_individual_outcomes(self, X, T, response='bin', soft=True):
        """Return individual treatment outcomes.
        
        Args:
            X (torch.Tensor): Covariate matrix of size (num_subjects, num_features).
            T (torch.Tensor): Treatment assignment of size (num_subjects,).
            response (str):
            soft (bool):

        Returns:
            list of torch.Tensor, one for each unique value in tensor T
        """
        output = []
        for t in torch.unique(T, sorted=True):
            o = self.forward(X, torch.full_like(T, t), response=response, soft=soft)
            output.append(o)
        return output
        # return [self.forward(X, torch.full_like(T, t), response=response, soft=soft)
        #         for t in torch.unique(T, sorted=True)]

    def get_groups(self, X):
        """Return hard assignment of groups for each sample

        Args:
            X (torch.Tensor): Covariate matrix of size (num_subjects, num_features).

        Returns:
            groups (np.ndarray): Most probable group assignment of each sample, size = (num_samples,)
        """
        # groups = self.forward(X, T, infer=False)[0].data.numpy()
        # groups = np.exp(groups) / np.exp(groups).sum(axis=1).reshape(-1, 1)
        groups = self.get_groups_proba(X)
        groups = np.argmax(groups, axis=1)
        return groups

    def get_groups_proba(self, X, log=False):
        """Return soft assignment of probability of each sample to be part of each group.

        Args:
            X (torch.Tensor): Covariate matrix of size (num_subjects, num_features).
            log (bool): If True returns log probabilities

        Returns:
            z_pred (np.array): probability of group membership given X P(Z|X),
                               size = (num_covariates, num_components+1).
        """
        a = torch.ones(X.shape[0], dtype=torch.float64)
        z_pred = self.forward(X, a, infer=False)[1].cpu().data.numpy()
        if not log:
            z_pred = np.exp(z_pred)
        return z_pred

    def get_groups_effect(self):
        group_effect = self.treat.cpu().data.numpy()
        return group_effect



import numpy as np
from sklearn.base import BaseEstimator, MetaEstimatorMixin, clone
from sklearn.utils.multiclass import type_of_target
from sklearn.linear_model import LassoCV, LogisticRegressionCV
# Find internal implementations
try:  # Version 0.20 - 0.21
    from sklearn.feature_selection.base import SelectorMixin
except ModuleNotFoundError:
    # Version >= 0.22
    from sklearn.feature_selection._base import SelectorMixin


__all__ = ["DoubleLASSO", "RecursiveConfounderElimination"]


# noinspection PyAbstractClass
class _BaseConfounderSelection(SelectorMixin, MetaEstimatorMixin, BaseEstimator):
    def __init__(self, importance_getter='auto', covariates=None):
        self.importance_getter = importance_getter
        self.covariates = covariates

    # TODO: move functionality for general utils and use for Estimators as well
    @staticmethod
    def _filter_covariates(func):
        def filter_covariates_and_run(self, X, *args, **kwargs):
            covariates = self.covariates
            if covariates is None:
                covariates = X.columns
            X = X.loc[:, covariates]
            return func(self, X, *args, **kwargs)
        return filter_covariates_and_run

    # @staticmethod
    def _filter_and_re_add_covariates(func):
        def filter_covariates_and_run_and_add(self, X, *args, **kwargs):
            covariates = self.covariates
            if covariates is None:
                covariates = X.columns
            sub_X = X.loc[:, covariates]
            res = func(self, sub_X, *args, **kwargs)
            complement_covariates = X.columns.difference(sub_X.columns)
            res = res.join(X.loc[:, complement_covariates])
            return res
        return filter_covariates_and_run_and_add

    @_filter_and_re_add_covariates
    def transform(self, X, a=None):
        X = X.loc[:, self.get_support()]
        return X

    # Decorator function cannot be defined as static before decorating,
    # so setting as the decorator as `staticmethod` is done after defining the `transform` using the decorator
    _filter_and_re_add_covariates = staticmethod(_filter_and_re_add_covariates)


class DoubleLASSO(_BaseConfounderSelection):
    def __init__(self, treatment_lasso=None, outcome_lasso=None,
                 mask_fn=None, threshold=1e-6,
                 importance_getter='auto', covariates=None):
        """
        A method for selecting confounders using sparse regression
        on both the treatment and the outcomes, and select for

        Implementing "Inference on Treatment Effects after Selection
        among High-Dimensional Controls"
        https://academic.oup.com/restud/article/81/2/608/1523757

        Args:
            treatment_lasso: Lasso learner to fit confounders and treatment.
                For example using scikit-learn,
                continuous treatment may use: `Lasso()`,
                discrete treatment may use: `LogisticRegression(penalty='l1')`.
                If `None` will try to automatically assign a lasso model with cross validation.
            outcome_lasso: Lasso learner to fit confounders and outcome.
                For example using scikit-learn,
                continuous outcome may use: `Lasso()`,
                discrete outcome may use: `LogisticRegression(penalty='l1')`.
                If `None` will try to automatically assign lasso model cross-validation.
            mask_fn: Function that takes input as two fitted lasso learners
                and returns a mask of the length of number of columns where True
                corresponds to columns that need to be selected. When set to None,
                the default implementation returns a mask based on non-zero
                coefficients in either learner. User can supply their own function,
                which must return a boolean array (of the length of columns of X)
                to indicate which columns are to be included.
            threshold: For default mask_fn, absolute value below which a lasso
                coefficient is treated as zero.
            importance_getter (str | callable): how to obtain feature importance.
                either a callable that inputs an estimator,
                a string of `'coef_'` or `'feature_importance_'`,
                or `'auto'` will detect `'coef_'` or `'feature_importance_'` automatically.
            covariates (list | np.ndarray): Specifying a subset of columns to perform selection on.
                Columns in `X` but not in `covariates` will be included after `transform`
                no matter the selection.
                Can be either a list of column names, or an array of boolean indicators length of `X`,
                or anything compatible with pandas `loc` function for columns.
                if `None` then all columns are participating in the selection process.
                This is similar to using sklearn's `ColumnTransformer` or `make_column_selector`.
        """
        # TODO: allowing users to provide the models follows the same design
        #       design principle throughout causallib,
        #       however, this might put some strain on the users who will need
        #       to know to supply `sklearn.linear_model.LogisticRegression(penalty='l1')
        #       for the treatment and the same for categorical outcome, but
        #       `sklearn.linear_model.Lasso` for continuous outcome.
        super().__init__(importance_getter, covariates)
        self.treatment_lasso = treatment_lasso
        self.outcome_lasso = outcome_lasso
        self.mask_fn = mask_fn
        self.threshold = threshold

    @_BaseConfounderSelection._filter_covariates
    def fit(self, X, a, y):
        self.treatment_lasso = self._data_driven_initialization(self.treatment_lasso, a)
        self.outcome_lasso = self._data_driven_initialization(self.outcome_lasso, y)

        self.treatment_lasso.fit(X, a)
        self.outcome_lasso.fit(X, y)
        mask_fn = self.mask_fn or self._get_non_zero_coef_mask
        self.support_ = mask_fn(self.treatment_lasso, self.outcome_lasso)
        self.n_features_ = self.support_.sum()
        return self

    # @_BaseConfounderSelection._filter_and_re_add_covariates
    # def transform(self, X, a=None):
    #     X = X.iloc[:, self.get_support()]
    #     return X

    def _get_support_mask(self):
        return self.support_

    def _get_non_zero_coef_mask(self, treatment_lasso, outcome_lasso):
        # Using _get_feature_importances from sklearn covers many more
        # edge cases than writing a vanilla function.
        # Specifying transform_func "norm" in the call below
        # actually calls np.abs when the coef_ attribute of treatment_lasso
        # (or output_lasso) is a one dimensional vector.
        treatment_lasso_importances = _get_feature_importances(
            treatment_lasso, self.importance_getter, transform_func="norm",
        )
        outcome_lasso_importances = _get_feature_importances(
            outcome_lasso, self.importance_getter, transform_func="norm",
        )
        treatment_mask = treatment_lasso_importances >= self.threshold
        outcome_mask = outcome_lasso_importances >= self.threshold
        return treatment_mask | outcome_mask

    @staticmethod
    def _data_driven_initialization(estimator, target):
        if estimator is not None:  # User provided an estimator
            return estimator

        if type_of_target(target) == "continuous":
            estimator = LassoCV()
        else:
            estimator = LogisticRegressionCV(penalty='l1', solver='saga', max_iter=5000)
        return estimator


class RecursiveConfounderElimination(_BaseConfounderSelection):

    def __init__(self, estimator, n_features_to_select: int = 1, step: int = 1,
                 importance_getter="auto", covariates=None):
        """Recursively eliminate confounders to prune confounders.

        Args:
            estimator: Estimator to fit for every step of recursive elimination.
            n_features_to_select (int): The number of confounders to keep.
            step (int): The number of confounders to eliminate in one iteration.
            importance_getter (str | callable): how to obtain feature importance.
              either a callable that inputs an estimator,
              a string of `'coef_'` or `'feature_importance_'`,
              or `'auto'` will detect `'coef_'` or `'feature_importance_'` automatically.
            covariates (list | np.ndarray): Specifying a subset of columns to perform selection on.
              Columns in `X` but not in `covariates` will be included after `transform`
              no matter the selection.
              Can be either a list of column names, or an array of boolean indicators length of `X`,
              or anything compatible with pandas `loc` function for columns.
              if `None` then all columns are participating in the selection process.
              This is similar to using sklearn's `ColumnTransformer` or `make_column_selector`.
        """
        super().__init__(importance_getter, covariates)
        self.estimator = estimator
        self.n_features_to_select = n_features_to_select
        self.step = step

    @_BaseConfounderSelection._filter_covariates
    def fit(self, X, a, y):
        # This is like an abbreviated implementation of RFE in sklearn.
        # Main differences are (a) Conditioning on treatment for every iteration,
        # (b) adjusting ranking/support not to include treatment, and (c) accounting
        # for causallib data types for X and a.
        # TODO: the entire implementation may be reduced to overwriting the
        #       `importance_getter` function and rigging it to have infinity
        #       importance for the treatment assignment every time.
        n_features = len(X.columns)
        support_ = np.ones(n_features, dtype=bool)
        ranking_ = np.ones(n_features, dtype=int)
        while np.sum(support_) > self.n_features_to_select:
            features = np.arange(n_features)[support_]
            estimator = clone(self.estimator)
            estimator.fit(a.to_frame().join(X.iloc[:, features]), y)
            importances = _get_feature_importances(
                estimator, self.importance_getter, transform_func="square",
            )
            importances = importances[1:]  # Do not consider "a" for dropping
            ranks = np.argsort(importances)
            ranks = np.ravel(ranks)
            threshold = min(self.step, np.sum(support_) - self.n_features_to_select)
            support_[features[ranks][:threshold]] = False
            ranking_[np.logical_not(support_)] += 1
        features = np.arange(n_features)[support_]
        self.estimator.fit(a.to_frame().join(X.iloc[:, features]), y)
        self.n_features_ = support_.sum()
        self.support_ = support_
        self.ranking_ = ranking_
        return self

    # @_BaseConfounderSelection._filter_and_re_add_covariates
    # def transform(self, X, a=None):
    #     X = X.iloc[:, self.get_support()]
    #     return X

    def _get_support_mask(self):
        return self.support_


def _get_feature_importances(estimator, getter, transform_func=None, norm_order=1):
    """
    Retrieve and aggregate (if ndim > 1) (and optionally transforms)
    the feature importances from an estimator.

    Args:
        estimator: A scikit-learn estimator from which we want to get the feature importances.
        getter (str | callable):  An attribute or a callable to get the feature importance.
                If `"auto"`, `estimator` is expected to expose `coef_` or `feature_importances_`.
        transform_func (str | None): The transform to apply to the feature importances.
                By default (`None`) no transformation is applied.
                Only "norm" and "square" are currently supported.
        norm_order (int): The norm order to apply when `transform_func="norm"`.
                Only applied when `importances.ndim > 1`.

    Returns:
        np.ndarray: The features importances, optionally transformed.
    """
    # A local version of sklearn's `_get_feature_importance`,
    # Because there have been multiple changes between version 0.20 and 0.24
    # in both API and import location, it uses the 0.24 version of the function.
    # Once dependencies move to > 0.24 this may be removed

    if isinstance(getter, str):
        if getter == "auto":
            if hasattr(estimator, "coef_"):
                getter = "coef_"
            elif hasattr(estimator, "feature_importances_"):
                getter = "feature_importances_"
            else:
                raise ValueError(
                    f"`importance_getter=='auto'` requires the estimator to have "
                    f"a `coef_` or `feature_importances_` attribute. "
                    f"If your estimator should have these attributes, "
                    f"make sure it is fitted before calling transform."
                )
        importances = getattr(estimator, getter)
    elif callable(getter):
        importances = getter(estimator)
    else:
        raise ValueError("`importance_getter` has to be a string or `callable`")

    if transform_func is None:
        pass
    elif transform_func == "norm":
        if importances.ndim == 1:
            importances = np.abs(importances)
        else:
            importances = np.linalg.norm(importances, axis=0, ord=norm_order)
    elif transform_func == "square":
        importances = importances ** 2
        if importances.ndim > 1:
            importances = importances.sum(axis=0)
    else:
        raise ValueError("`transform_func` only supports None, 'norm' and 'square'.")

    return importances



"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Apr 16, 2018


A module defining the various hierarchy of causal models interface.
Causal models have two main tasks - predicting counterfactual outcomes and predicting effect based on these estimated
outcomes.
On top of it there are two resolutions we can work on: the individual level (i.e. outcome and effect for each individual
in the dataset) and population level (i.e. some aggregation on the sample level).
This module defines it all with:
* EffectEstimator - can estimate both individual and population level effect
* PopulationOutcomeEstimator - estimates aggregated outcomes on different sub-groups in the dataset.
* IndividualOutcomeEstimator - estimates individual level outcomes.
"""
import abc
import warnings

import pandas as pd
from numpy import isscalar

from sklearn.base import BaseEstimator

from ..utils.general_tools import create_repr_string


class EffectEstimator(BaseEstimator):
    """
    Class-based interface for estimating either individual-level or sample-level effect.
    """
    # This is somewhat static class, but since we wish it be inherited interface, we grouped it as class rather than a
    # module.

    # Invariant to vector or scalar arithmetic:
    CALCULATE_EFFECT = {"diff": lambda x, y: x - y,
                        "ratio": lambda x, y: x / y,
                        "or": lambda x, y: (x / (1 - x)) / (y / (1 - y))
                        }

    def estimate_effect(self, outcome_1, outcome_2, effect_types="diff"):
        """
        Estimates an effect given two potential outcomes.

        Args:
            outcome_1 (pd.Series | float): A potential outcome.
            outcome_2 (pd.Series | float): A potential outcome.
            effect_types (list[str] | str): Any iterable of strings from the set of EffectEstimator.CALCULATE_EFFECT keys

        Returns:
            pd.Series | pd.DataFrame: A Series if population effect (input is scalar) with index are the effect types
                                      and values are the corresponding computed effect.
                                      A DataFrame if individual effect (input is a vector) where columns are effects
                                      types and rows are effect in each individual.
                                      Always: Value type is same is outcome_1 and outcome_2 type.
        Examples:
            >>> from causallib.estimation.base_estimator import EffectEstimator
            >>> effect_estimator = EffectEstimator()
            >>> effect_estimator.estimate_effect(0.3, 0.6)
            >>> {"diff": -0.3,    # 0.3 - 0.6
                 "ratio": 0.5,    # 0.3 / 0.6
                 "or": 0.2857}    # Odds-Ratio(0.3, 0.6)
        """
        effect_types = [effect_types] if isscalar(effect_types) else effect_types
        results = {}
        for effect_type in effect_types:
            effect = self.CALCULATE_EFFECT[effect_type](outcome_1, outcome_2)
            results[effect_type] = effect
        # Format results in pandas array:
        results = pd.Series(results) if isscalar(outcome_1) else pd.concat(results, axis="columns",
                                                                           names=["effect_type"])
        return results


class PopulationOutcomeEstimator(EffectEstimator):
    """
    Interface for estimating aggregated outcome over different subgroups in the dataset.
    """

    # def __init__(self, *args, **kwargs):
    #     super(PopulationOutcomeEstimator, self).__init__(*args, **kwargs)

    @abc.abstractmethod
    def estimate_population_outcome(self, X, a, y, treatment_values=None):
        raise NotImplementedError


class IndividualOutcomeEstimator(PopulationOutcomeEstimator, EffectEstimator):
    """
    Interface for estimating individual-level outcome for different treatment values.
    """

    def __init__(self, learner, predict_proba=False, *args, **kwargs):
        """

        Args:
            learner: Initialized sklearn model.
            predict_proba (bool): In case the outcome task is classification and in case `learner` supports the
                                  operation, if True - prediction will utilize learner's `predict_proba` or
                                  `decision_function` which returns a continuous matrix of size (n_samples, n_classes).
                                  If False - `predict` will be used and return value will be based on a vector of class
                                  classifications.
        """
        self.learner = learner
        self.predict_proba = predict_proba
        # super(IndividualOutcomeEstimator, self).__init__(*args, **kwargs)

    @staticmethod
    def _aggregate_population_outcome(y, agg_func="mean"):
        """
        Aggregates a vector (of individual outcomes) to a population outcome scalar.

        Args:
            y (pd.Series): Individual outcome prediction.
            agg_func (str): Type of aggregation function (e.g. "mean" or "median").

        Returns:
            float: A scalar (float) aggregated result on the input.
        """
        if agg_func == "mean":
            return y.mean()
        elif agg_func == "median":
            return y.median()
        # TODO: consider adding max and min aggregation
        else:
            raise LookupError("Not supported aggregation function ({})".format(agg_func))

    def estimate_population_outcome(self, X, a, y=None, treatment_values=None, agg_func="mean"):
        """
        Implements aggregation of individual outcome into population (sample) outcome.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            y (pd.Series | None): Observed outcome of size (num_subjects,).
            treatment_values (Any): Desired treatment value/s to stratify upon before aggregating individual into
                                    population outcome.
                                    If not supplied, calculates for all available treatment values.
            agg_func (str): Type of aggregation function (e.g. "mean" or "median").

        Returns:
            pd.Series: Series which index are treatment values, and the values are numbers - the aggregated outcome for
                       the strata of people whose assigned treatment is the key.
        """
        if y is not None:
            warnings.warn("Variable y (outcome) is not used when calculating population outcome of an "
                          "IndividualOutcome estimator. Instead it utilizes the individual outcome estimation.")
        individual_cf = self.estimate_individual_outcome(X, a, treatment_values)
        pop_outcome = individual_cf.apply(self._aggregate_population_outcome, args=(agg_func,))
        return pop_outcome

    def estimate_effect(self, outcome1, outcome2, agg="population", effect_types="diff"):
        """
        Estimates an effect given two potential outcomes.

        Args:
            outcome1 (pd.Series): A potential outcome.
            outcome2 (pd.Series): A potential outcome.
            agg (str): Either "population" or "individual" - whether to calculate individual effect or population
                       effect.
            effect_types (list[str] | str): Any iterable of strings from the set of EffectEstimator.CALCULATE_EFFECT keys

        Returns:
            pd.Series | pd.DataFrame: A Series if population effect (input is scalar) with index are the effect types
                                      and values are the corresponding computed effect.
                                      A DataFrame if individual effect (input is a vector) where columns are effects
                                      types and rows are effect in each individual.
                                      Always: Value type is the same as outcome_1 and outcome_2 type.
        """
        if agg == "population":
            outcome1 = self._aggregate_population_outcome(outcome1)
            outcome2 = self._aggregate_population_outcome(outcome2)
        effect = super(IndividualOutcomeEstimator, self).estimate_effect(outcome1, outcome2, effect_types)
        return effect

    @abc.abstractmethod
    def estimate_individual_outcome(self, X, a, treatment_values=None, predict_proba=None):
        """
        Estimates individual outcome under different treatment values (interventions)

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            treatment_values (Any): Desired treatment value/s to use when estimating the counterfactual outcome/
                                    If not supplied, calculates for all available treatment values.
            predict_proba (bool | None): In case the outcome task is classification and in case `learner` supports the
                                         operation, if True - prediction will utilize learner's `predict_proba` or
                                         `decision_function` which returns a continuous matrix of size
                                         (n_samples, n_classes).
                                         If False - `predict` will be used and return value will be based on a vector of
                                         class classifications.
                                         If None - parameter is ignored and behaviour is as specified when initializing
                                         the IndividualOutcomeEstimator.

        Returns:
            pd.DataFrame: DataFrame which columns are treatment values and rows are individuals: each column is a vector
                          size (num_samples,) that contains the estimated outcome for each individual under the
                          treatment value in the corresponding key.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def fit(self, X, a, y, sample_weight=None):
        """
        Trains a causal model from observed data.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            y (pd.Series): Observed outcome of size (num_subjects,).
            sample_weight: To be passed to the underlining scikit-learn's fit method.

        Returns:
            IndividualOutcomeEstimator: A causal weight model with an inner learner fitted.
        """
        raise NotImplementedError

    # @abc.abstractmethod
    # def predict(self, X, a):
    #     raise NotImplementedError

    def evaluate_fit(self, X, y, a=None):
        # if a is given then you can return fit on subgroups stratified by treatment values
        pass  # TODO: Implement

    def __repr__(self):
        repr_string = create_repr_string(self)
        return repr_string



# (C) Copyright 2020 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Created on Nov 12, 2020
import unittest

import matplotlib
import matplotlib.axes
import numpy as np
import pandas as pd

from sklearn.linear_model import LogisticRegression, LinearRegression

from causallib.evaluation import evaluate
from causallib.evaluation.evaluator import evaluate_bootstrap
from causallib.evaluation.metrics import (
    get_default_binary_metrics,
    get_default_regression_metrics,
)
from causallib.estimation import AIPW, IPW, StratifiedStandardization
from causallib.datasets import load_nhefs


matplotlib.use("Agg")


def binarize(cts_output: pd.Series) -> pd.Series:
    """Turn continuous outcome into binary by applying sigmoid.

    Args:
        cts_output (pd.Series): outcomes as continuous variables

    Returns:
        pd.Series: outcomes as binary variables
    """

    y = 1 / (1 + np.exp(-cts_output))
    y = np.random.binomial(1, y)
    y = pd.Series(y, index=cts_output.index)
    return y


class TestEvaluations(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        data = load_nhefs()
        self.X, self.a, self.y = data.X, data.a, data.y
        self.y_bin = binarize(data.y)
        ipw = IPW(LogisticRegression(solver="liblinear"), clip_min=0.05, clip_max=0.95)
        std = StratifiedStandardization(LinearRegression())
        self.dr = AIPW(std, ipw)
        self.dr.fit(self.X, self.a, self.y)
        self.std_bin = StratifiedStandardization(LogisticRegression(solver="liblinear"))
        self.std_bin.fit(self.X, self.a, self.y_bin)

    def test_evaluate_bootstrap_with_refit_works(self):
        ipw = IPW(LogisticRegression(solver="liblinear"), clip_min=0.05, clip_max=0.95)
        evaluate_bootstrap(ipw, self.X, self.a, self.y, n_bootstrap=5, refit=True)

    def test_evaluate_cv_works_with_unfit_models(self):
        ipw = IPW(LogisticRegression(solver="liblinear"), clip_min=0.05, clip_max=0.95)
        evaluate(ipw, self.X, self.a, self.y, cv="auto")

    def test_metrics_to_evaluate_is_none_means_no_metrics_evaluated(self):
        for model in (self.dr.outcome_model, self.dr.weight_model):
            self.ensure_metrics_are_none(model)

    def ensure_metrics_are_none(self, model):
        results = evaluate(model, self.X, self.a, self.y, metrics_to_evaluate=None)
        self.assertIsNone(results.evaluated_metrics)

    def test_default_evaluation_metrics_weights(self):
        model = self.dr.weight_model
        results = evaluate(model, self.X, self.a, self.y)
        self.assertEqual(
            set(results.evaluated_metrics.prediction_scores.columns),
            set(get_default_binary_metrics().keys()),
        )

    def test_default_evaluation_metrics_continuous_outcome(self):
        model = self.dr.outcome_model
        results = evaluate(model, self.X, self.a, self.y)
        self.assertEqual(
            set(results.evaluated_metrics.columns),
            set(get_default_regression_metrics().keys()),
        )

    def test_default_evaluation_metrics_binary_outcome(self):
        model = self.std_bin
        results = evaluate(model, self.X, self.a, self.y_bin)
        self.assertEqual(
            set(results.evaluated_metrics.columns),
            set(get_default_binary_metrics().keys()),
        )



from itertools import product

import numpy as np
import pandas as pd

from sklearn.model_selection import StratifiedKFold
from sklearn.utils.multiclass import type_of_target


class TreatmentOutcomeStratifiedKFold(StratifiedKFold):
    """Creates stratified folds based on both the treatment assignment 
    and the outcome. 
    That is, every fold preserves both the treatment prevalence and
    outcome prevalence within each treatment. 
    
    For non-class outcomes, stratification is done only based on treatment. 
    
    """
    __doc__ += StratifiedKFold.__doc__

    @staticmethod
    def _combine_treatment_outcome_labels(a, y):
        """combines every `a` x `y` values as a unique label"""
        # Assuming n_a < 10, n_y < 10: labels = a*10+y. Implements a generic version.
        a_unique = np.unique(a)
        y_unique = np.unique(y)
        combinations = product(a_unique, y_unique)
        combinations_mapping = {c: i for i, c in enumerate(combinations)}
        combined_labels = [combinations_mapping[(ai, yi)] for ai, yi in zip(a, y)]
        combined_labels = pd.Series(combined_labels, index=a.index)
        return combined_labels

    def _get_labels_for_split(self, a, y):
        target_type = type_of_target(y)
        if target_type not in ("binary", "multiclass"):
            # `y` is incompatible with stratification
            raise ValueError(
                f"Outcome type should either be 'binary' or 'multiclass'."
                f"Received {target_type} instead."
            )
        labels = self._combine_treatment_outcome_labels(a, y)
        return labels

    def split(self, joinedXa, y, groups=None):
        X = joinedXa.iloc[:, :-1]
        a = joinedXa.iloc[:, -1]
        splits = self._split(X, a, y, groups=groups)
        # labels = self._get_labels_for_split(a, y)
        # splits = super().split(X, labels, groups=groups)
        return splits

    def _split(self, X, a, y, groups=None):
        """A causallib-like `X, a, y` interface for split"""
        labels = self._get_labels_for_split(a, y)
        splits = super().split(X, labels, groups=groups)
        return splits


class TreatmentStratifiedKFold(StratifiedKFold):
    """Creates stratified folds based on the treatment assignment.
    That is, every fold preserves the treatment prevalence.
    """
    __doc__ += StratifiedKFold.__doc__

    def split(self, joinedXa, y=None, groups=None):
        X = joinedXa.iloc[:, :-1]
        a = joinedXa.iloc[:, -1]
        splits = self._split(X, a, y, groups=groups)
        return splits

    def _split(self, X, a, y=None, groups=None):
        """A causallib-like `X, a, y` interface for split"""
        splits = super().split(X, a, groups=groups)
        return splits



# (C) Copyright 2020 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Created on Nov 12, 2020
import unittest

import matplotlib
import matplotlib.axes
import matplotlib.pyplot as plt

from sklearn.linear_model import LogisticRegression, LinearRegression
import numpy as np
import pandas as pd


from causallib.evaluation import evaluate
from causallib.estimation import AIPW, IPW, StratifiedStandardization
from causallib.datasets import load_nhefs


matplotlib.use("Agg")


def binarize(cts_output: pd.Series) -> pd.Series:
    """Turn continuous outcome into binary by applying sigmoid.

    Args:
        cts_output (pd.Series): outcomes as continuous variables

    Returns:
        pd.Series: outcomes as binary variables
    """

    y = 1 / (1 + np.exp(-cts_output))
    y = np.random.binomial(1, y)
    y = pd.Series(y, index=cts_output.index)
    return y


class TestPlots(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        data = load_nhefs()
        self.X, self.a, self.y = data.X, data.a, data.y
        ipw = IPW(LogisticRegression(solver="liblinear"), clip_min=0.05, clip_max=0.95)
        std = StratifiedStandardization(LinearRegression())
        self.dr = AIPW(std, ipw)
        self.dr.fit(self.X, self.a, self.y)
        self.cts_outcome_evaluation = evaluate(
            self.dr.outcome_model, self.X, self.a, self.y
        )
        self.propensity_evaluation = evaluate(self.dr.weight_model, self.X, self.a, self.y)
        self.y_bin = binarize(data.y)
        self.std_bin = StratifiedStandardization(LogisticRegression(solver="liblinear"))
        self.std_bin.fit(self.X, self.a, self.y_bin)
        self.bin_outcome_evaluation = evaluate(self.std_bin, self.X, self.a, self.y_bin)

    def test_propensity_plots_only_exist_for_propensity_model(self):
        self.propensity_evaluation.plot_covariate_balance
        self.propensity_evaluation.plot_weight_distribution
        self.propensity_evaluation.plot_pr_curve
        self.propensity_evaluation.plot_roc_curve
        self.propensity_evaluation.plot_calibration_curve

        with self.assertRaises(AttributeError):
            self.cts_outcome_evaluation.plot_covariate_balance
            self.cts_outcome_evaluation.plot_weight_distribution
            self.cts_outcome_evaluation.plot_pr_curve
            self.cts_outcome_evaluation.plot_roc_curve
            self.cts_outcome_evaluation.plot_calibration_curve

    def test_outcome_plots_only_exist_for_outcome_model(self):
        self.cts_outcome_evaluation.plot_continuous_accuracy
        self.cts_outcome_evaluation.plot_residuals
        self.cts_outcome_evaluation.plot_common_support

        with self.assertRaises(AttributeError):
            self.propensity_evaluation.plot_continuous_accuracy
            self.propensity_evaluation.plot_residuals
            self.propensity_evaluation.plot_common_support

    def test_weight_distribution_reflect_has_negative_yaxis(self):
        f, ax = plt.subplots()
        axis = self.propensity_evaluation.plot_weight_distribution(reflect=True, ax=ax)
        self.assertIsInstance(axis, matplotlib.axes.Axes)
        minx, maxx, miny, maxy = axis.axis()
        self.assertLess(miny, 0)
        self.assertLess(maxx, 1)
        self.assertGreater(minx, 0)
        self.assertGreater(maxy, 0)
        plt.close()

    def test_weight_distribution_noreflect_has_nonegative_yaxis(self):
        f, ax = plt.subplots()
        axis = self.propensity_evaluation.plot_weight_distribution(reflect=False, ax=ax)
        self.assertIsInstance(axis, matplotlib.axes.Axes)
        minx, maxx, miny, maxy = axis.axis()
        self.assertEqual(miny, 0)
        self.assertLess(maxx, 1)
        self.assertGreater(minx, 0)
        self.assertGreater(maxy, 0)
        plt.close()

    def test_plot_covariate_balance_love_draws_thresh(self):
        thresh = 0.1
        f, ax = plt.subplots()
        axis = self.propensity_evaluation.plot_covariate_balance(
            kind="love", thresh=thresh, ax=ax
        )
        self.assertIsInstance(axis, matplotlib.axes.Axes)
        self.assertEqual(thresh, axis.get_lines()[0].get_xdata()[0])
        plt.close()


    def test_plot_covariate_balance_scatter_draws_thresh(self):
        thresh = 0.1
        f, ax = plt.subplots()
        axis = self.propensity_evaluation.plot_covariate_balance(
            kind="scatter", thresh=thresh, ax=ax
        )
        self.assertIsInstance(axis, matplotlib.axes.Axes)
        self.assertEqual(thresh, axis.get_lines()[0].get_xdata()[0])
        plt.close()

    def test_plot_covariate_balance_slope_labeled_correctly(self):
        f, ax = plt.subplots()
        axis = self.propensity_evaluation.plot_covariate_balance(kind="slope", ax=ax)
        self.assertIsInstance(axis, matplotlib.axes.Axes)
        self.assertEqual([x.get_xdata() for x in axis.get_lines()][1][0], "unweighted")
        plt.close()

    def test_roc_curve_has_dashed_diag(self):
        self.ensure_roc_curve_has_dashed_diag(self.propensity_evaluation)
        self.ensure_roc_curve_has_dashed_diag(self.bin_outcome_evaluation)

    def test_calibration_curve_has_dashed_diag(self):
        self.ensure_calibration_curve_has_dashed_diag(self.propensity_evaluation)
        self.ensure_calibration_curve_has_dashed_diag(self.bin_outcome_evaluation)

    def test_pr_curve_has_flat_dashed_chance_line(self):
        self.ensure_pr_curve_has_flat_dashed_chance_line(
            self.propensity_evaluation, chance=self.a.mean()
        )
        self.ensure_pr_curve_has_flat_dashed_chance_line(
            self.bin_outcome_evaluation, chance=self.y_bin.mean()
        )

    def test_accuracy_plot_has_dashed_diag(self):
        f, ax = plt.subplots()
        axis = self.cts_outcome_evaluation.plot_continuous_accuracy(ax=ax)
        self.assertIsInstance(axis, matplotlib.axes.Axes)
        diag = [x for x in axis.lines if len(x.get_xdata()) == 2][0]
        self.assertEqual(diag.get_linestyle(), "--")
        self.assertTrue(all(diag.get_xdata() == diag.get_ydata()))
        plt.close()

    def test_common_support_plot_has_dashed_diag(self):
        f, ax = plt.subplots()
        axis = self.cts_outcome_evaluation.plot_common_support(ax=ax)
        self.assertIsInstance(axis, matplotlib.axes.Axes)
        diag = [x for x in axis.lines if len(x.get_xdata()) == 2][0]
        self.assertEqual(diag.get_linestyle(), "--")
        self.assertTrue(all(diag.get_xdata() == diag.get_ydata()))
        plt.close()
        plt.close()

    def test_plot_all_generates_correct_plot_names(self):
        self.ensure_plot_all_generates_all_plot_names(self.propensity_evaluation)
        self.ensure_plot_all_generates_all_plot_names(self.cts_outcome_evaluation)
        self.ensure_plot_all_generates_all_plot_names(self.bin_outcome_evaluation)



    def test_residuals_plot_has_dashed_zero_line(self):
        f, ax = plt.subplots()
        axis = self.cts_outcome_evaluation.plot_residuals(ax=ax)
        self.assertIsInstance(axis, matplotlib.axes.Axes)
        zero_line = [x for x in axis.lines if len(x.get_xdata()) == 2][0]
        self.assertEqual(zero_line.get_linestyle(), "--")
        self.assertEqual(zero_line.get_ydata()[0], 0)
        self.assertEqual(zero_line.get_ydata()[1], 0)
        plt.close()

    def ensure_roc_curve_has_dashed_diag(self, results_object):
        f, ax = plt.subplots()
        axis = results_object.plot_roc_curve(ax=ax)
        self.assertIsInstance(axis, matplotlib.axes.Axes)
        diag = [x for x in axis.lines if len(x.get_xdata()) == 2][0]
        self.assertEqual(diag.get_linestyle(), "--")
        self.assertTrue(all(diag.get_xdata() == [0, 1]))
        self.assertTrue(all(diag.get_ydata() == [0, 1]))
        plt.close()

    def ensure_calibration_curve_has_dashed_diag(self, results_object):
        f, ax = plt.subplots()
        axis = results_object.plot_calibration_curve(ax=ax)
        self.assertIsInstance(axis, matplotlib.axes.Axes)
        diag = [x for x in axis.lines if len(x.get_xdata()) == 2][0]
        self.assertEqual(diag.get_linestyle(), "--")
        self.assertTrue(all(diag.get_xdata() == diag.get_ydata()))
        plt.close()

    def ensure_pr_curve_has_flat_dashed_chance_line(self, results_object, chance):
        f, ax = plt.subplots()
        axis = results_object.plot_pr_curve(ax=ax)
        self.assertIsInstance(axis, matplotlib.axes.Axes)
        chance_line = [x for x in axis.lines if len(x.get_xdata()) == 2][0]
        self.assertEqual(chance_line.get_label(), "Chance")
        self.assertEqual(chance_line.get_linestyle(), "--")
        self.assertAlmostEqual(chance_line.get_ydata()[0], chance)
        self.assertAlmostEqual(chance_line.get_ydata()[1], chance)
        plt.close()

    def ensure_plot_all_generates_all_plot_names(self, evaluation_results):
        all_plots = evaluation_results.plot_all()
        all_names = evaluation_results.all_plot_names
        self.assertEqual(set(all_plots["train"].keys()), all_names)


# (C) Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Created on Oct 30, 2019

from sklearn.preprocessing import LabelEncoder
from sklearn.utils.class_weight import compute_class_weight

from .classifier_selection import select_classifier
from ...estimation.base_weight import WeightEstimator, PropensityEstimator
from ...estimation.base_estimator import PopulationOutcomeEstimator

import numpy as np
import pandas as pd
import copy


class AdversarialBalancing(WeightEstimator, PopulationOutcomeEstimator):
    def __init__(self, learner, iterations=20, lr=0.5, decay=1, loss_type='01', use_stabilized=True,
                 verbose=False, *args, **kwargs):
        """
        Adversarial Balancing finds sample weights such that the weighted population under any treatment A
        looks similar (distribution-wise) to the true population.
        Borrowing from GANs, the main idea is that, for each treatment A, the algorithm find weights such that
        a specified classifier cannot distinguish between the entire population and the weighted population under
        treatment `a`.

        At each step we update the weights using the gradient of the exponential loss function, and re-train the
        classifier.
        For a given classifier family, an optimal solution are weights that maximize the minimal error of classifiers
        in this family.

        For more details about the algorithm see:
         *Adversarial Balancing for Causal Inference* by Ozery-Flato and Thodoroff et al.
         https://arxiv.org/abs/1810.07406

        Args:
            learner: An initialized classifier object implementing fit and predict (scikit-learn compatible)
                     Will be used to discriminate between the population under treatment a and the entire
                     global population.
                     A selection for each treatment value can be performed to choose the best classifier for that
                     treatment group. It can be done by providing a scikit-learn initialized SearchCV model (either
                     GridSearchCV or RandomizedSearchCV), or by providing a list of classifiers.
                     If providing a list of classifiers, a selection will be done for each treatment value using
                     cross-validation that will use the best-performing classifier among the list.
                     see select_classifier module.
            iterations (int): The number of iterations to adjust the weights of each sample
            lr (float): Learning rate used to update the weights
            decay (float):  Parameter to decay the learning rate through the iterations
            loss_type (str): Use '01' for zero-one loss, otherwise cross-entropy is used (and provided `learner` should
                             also implement `predict_proba` methods).
            use_stabilized (bool): Whether to re-weigh the learned weights with the prevalence of the treatment.
                                   Note: Adversarial balancing already has inherent component weighting treatment
                                    prevalence.
                                    Setting to False will "de-stabilize" the weights after they are calculated.
            verbose (bool): Whether to print out statistics to console during training.

        Attributes:
            iterative_models_: np.ndarray of size(n_treatment_values, iterations) holding all the models created
                               during training process.
            iterative_normalizing_consts_: np.ndarray of size(n_treatment_values, iterations) holding all the
                                           normalizing constants calculated during training process.
            discriminator_loss_: np.ndarray of size(n_treatment_values, iterations) holding the loss of the learner
                                 throughout the training process.
            treatments_frequency_: if use_stabilized=True, the proportions of the treatment values.
        """
        super().__init__(learner, use_stabilized, *args, **kwargs)  # Crashes when provided a list of classifiers

        self.iterations = iterations
        self.loss_type = loss_type
        self.lr = lr
        self.decay = decay
        self.verbose = verbose
        self.use_stabilized = use_stabilized

    def fit(self, X, a, y=None, w_init=None, **select_kwargs):
        """
        Trains an Adversarial Balancing model.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            y: IGNORED.
            w_init (pd.Series): Initial sample weights. If not provided, assumes uniform.
            select_kwargs: keywords argument to past into select_classifier.
                           relevant only if model was initialized with list of classifiers in `learner`.

        Returns:
            AdversarialBalancing
        """
        X = _to_ndarray(X)
        a = _to_ndarray(a)
        self._run(X, a, w_init=w_init, is_train=True, **select_kwargs)
        return self

    def compute_weights(self, X, a, treatment_values=None, use_stabilized=None, **kwargs):
        if treatment_values is not None:
            # Create synthetic treatment assignment of treatment_values instead of actual assignment `a`
            a = pd.Series(data=treatment_values, index=X.index)
        idx = X.index
        X, a = _to_ndarray(X), _to_ndarray(a)
        w = self._run(X, a, w_init=None, use_stabilized=use_stabilized, is_train=False)
        w = pd.Series(w, index=idx)
        return w

    def estimate_population_outcome(self, X, a, y, w=None, treatment_values=None):
        if w is None:
            w = self.compute_weights(X, a)
        res = self._compute_stratified_weighted_aggregate(y, sample_weight=w, stratify_by=a,
                                                          treatment_values=treatment_values)
        return res

    def _run(self, X, A, w_init=None, is_train=True, use_stabilized=None, **select_kwargs):
        if use_stabilized is None:
            use_stabilized = self.use_stabilized

        if w_init is None:
            w_init = np.ones((X.shape[0]))

        w = w_init.copy()  # Weights to be returned

        unique_treatments = np.sort(np.unique(A).astype(int))
        n_treatments = unique_treatments.shape[0]

        if is_train:
            if not np.all(unique_treatments == np.arange(n_treatments)):
                raise AssertionError("Treatment values in `a` must be indexed 0, 1, 2, ...")
            self.iterative_models_ = np.empty((n_treatments, self.iterations), dtype=object)
            self.iterative_normalizing_consts_ = np.full((n_treatments, self.iterations), np.NaN)

            self.discriminator_loss_ = np.zeros((n_treatments, self.iterations))
            self.treatments_frequency_ = _compute_treatments_frequency(A)

        for a in unique_treatments:
            # Create an artificial classification problem where the samples with label 1 are the original entire
            # population ("source population"),
            # and the samples with label -1 are the population under treatment a ("target population").
            # Labels 1 and -1 (rather than 0) are used because of the later exponential loss function
            X_augm = np.row_stack((X, X[A == a]))  # create the augmented dataset
            y = np.ones((X_augm.shape[0]))
            y[X.shape[0]:] *= -1  # subpopulation of current treatment (a) has y== -1
            target_pop_mask = y == -1

            # To simplify the task (learning weights) we ensure both target and source populations have the same
            # importance by reweighting classes by their frequency
            y_0_1 = LabelEncoder().fit_transform(y)  # Encode -1 ==> 0  and  1 ==>1
            class_weight = compute_class_weight(class_weight='balanced', classes=np.unique(y), y=y)[y_0_1]

            sample_weight = np.ones((X_augm.shape[0]))
            sample_weight[target_pop_mask] = w[A == a]  # Weights from initialization

            if is_train:
                selected_model = select_classifier(self.learner, X_augm, y,
                                                   loss_type=self.loss_type, **select_kwargs)

            for i in range(self.iterations):
                lr = self.lr * (1.0 / (1 + (self.decay * i)))  # decay the learning rate

                # fit model
                if is_train:
                    train_weight = sample_weight * class_weight
                    selected_model.fit(X_augm, y, sample_weight=train_weight)
                    self.iterative_models_[a, i] = copy.deepcopy(selected_model)
                else:
                    selected_model = self.iterative_models_[a, i]

                # get predictions and their errors
                pred, pred_loss = self.__get_predictions(selected_model, X_augm, y)

                if is_train:
                    total_error = np.sum(train_weight * pred_loss) / np.sum(train_weight)
                    # when using 0-1 loss:
                    # note that often np.sum(train_weight * pred_loss) != np.sum(train_weight[pred_loss.astype(bool)])
                    # and the difference is slight, probably due to numerical issues
                    self.discriminator_loss_[a, i] = total_error
                    if self.verbose:
                        print("Iteration {}/{} - loss: {}".format(i, self.iterations, total_error))

                # Update the weights to minimize the loss of the generator and "fool" the classifier
                if self.loss_type == '01':
                    sample_weight[target_pop_mask] *= np.exp(-lr * y[target_pop_mask] * pred[target_pop_mask])
                else:
                    sample_weight[target_pop_mask] *= np.exp(lr * pred_loss[target_pop_mask])

                if is_train:
                    tx_group_size = np.sum(target_pop_mask)
                    self.iterative_normalizing_consts_[a, i] = tx_group_size / np.sum(sample_weight[target_pop_mask])

                # Normalize the weights with mean 1 (so that weights.sum() == original group size)
                sample_weight[target_pop_mask] *= self.iterative_normalizing_consts_[a, i]

            w[A == a] = sample_weight[target_pop_mask]
            if not use_stabilized:  # Undo scaling by group size
                w[A == a] *= 1 / self.treatments_frequency_[a]

        return w

    def __get_predictions(self, selected_model, X_augm, y):
        if self.loss_type == '01':
            pred = selected_model.predict(X_augm).reshape((-1))
            pred_loss = (pred != y).astype(float)

        else:
            pred = selected_model.predict_proba(X_augm)[:, 1]
            pred_loss = np.log(pred[y == -1])
            pred_loss[y == 1] = np.log(1.0 - pred[y == 1])

        return pred, pred_loss

    def compute_weight_matrix(self, X, a, use_stabilized=None, **kwargs):
        res = {}
        for tx_val in np.sort(np.unique(a)):
            cur_assignment = pd.Series(data=tx_val, index=X.index)
            res[tx_val] = self.compute_weights(X, cur_assignment, use_stabilized=use_stabilized)
        res = pd.DataFrame(res)
        return res

    # def compute_propensity(self, X, a, treatment_values=None, **kwargs):
    #     """Computes probability for being treated
    #     (or probability of any treatment value provided in `treatment_values`).
    #
    #     Note that Adversarial Balance does not optimize for probabilities directly and
    #     this probabilities are calculated post-hoc from the weights.
    #
    #     Args:
    #         X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
    #         a (pd.Series): Treatment assignment of size (num_subjects,).
    #         treatment_values (Any | None): A desired value/s to extract propensity to (i.e. probabilities to what
    #                                       treatment value should be calculated).
    #                                       If not specified, then the maximal treatment value is chosen. This is since
    #                                       the usual case is of treatment (A=1) control (A=0) setting.
    #
    #     Returns:
    #         pd.Series: A vector size (num_subjects,) containing the propensity of each individual to be treatment
    #                    (or get the treatment value provided in `treatment_values`).
    #     """
    #     treatment_values = a.max() if treatment_values is None else treatment_values
    #     res = self.compute_propensity_matrix(X, a, use_stabilized=False)
    #     res = res[treatment_values]
    #     return res
    #
    # def compute_propensity_matrix(self, X, a=None, **kwargs):
    #     """Computes probability for being assigned to any treatment groups.
    #
    #     Note that Adversarial Balance does not optimize for probabilities directly and
    #     this probabilities are calculated post-hoc from the weights.
    #     Therefore, there's no guarantee all probabilities of a given sample will sum to 1.
    #
    #     Args:
    #         X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
    #         a (pd.Series | None): Treatment assignment of size (num_subjects,),
    #                               used to obtain the unique treatment values.
    #                               If not provided, uses all the treatment values seen during training.
    #
    #     Returns:
    #         pd.DataFrame: A matrix of size (num_subjects, num_treatments) with probability for every individual and e
    #                       very treatment.
    #     """
    #     if a is None:
    #         a = list(self.treatments_frequency_.keys())  # unique treatment values
    #     res = self.compute_weight_matrix(X, a, use_stabilized=False)
    #     res = 1 / res  # type: pd.DataFrame
    #     return res


def _compute_treatments_frequency(A):
    unique_treatments, unique_treatment_counts = np.unique(A, return_counts=True)
    unique_treatments_freq = unique_treatment_counts / (1.0*np.sum(unique_treatment_counts))
    treatments_frequency = dict(zip(unique_treatments, unique_treatments_freq))
    return treatments_frequency


def _to_ndarray(x):
    if not isinstance(x, np.ndarray):
        return np.array(x)
    else:
        return x



"""Scoring functions that operate on the evaluation results objects.

These functions depend on the causallib.evalutation results objects and are
less reusable than the functions in metrics.py.
"""
import pandas as pd

from .predictions import (
    OutcomePredictions,
    PropensityPredictions,
    WeightPredictions,
    WeightEvaluatorScores,
)

from .metrics import get_default_binary_metrics, get_default_regression_metrics

def score_cv(predictions, X, a, y, cv, metrics_to_evaluate="defaults"):
    """Evaluate the prediction against the true data using evaluation score metrics.

    Args:
        predictions (dict[str, list]): the output of predict_cv.
        X (pd.DataFrame): Covariates.
        a (pd.Series): Treatment assignment.
        y (pd.Series): Outcome.
        cv (list[tuples]): list the number of folds containing tuples of indices:
            (train_idx, validation_idx)
        metrics_to_evaluate (dict | "defaults"): key: metric's name, value: callable that receives
            true labels, prediction and sample_weights (the latter is allowed to be ignored).
            If `"defaults"`, default metrics are selected.
    Returns:
        pd.DataFrame | WeightEvaluatorScores:
            DataFrame whose columns are different metrics and each row is a
            product of phase x fold x strata.
            WeightEvaluatorScores also has a covariate-balance result in a DataFrame.
    """
    if metrics_to_evaluate == "defaults":
        metrics_to_evaluate = _get_default_metrics_to_evaluate(predictions["train"][0])

    phases = predictions.keys()
    scores = {phase: [] for phase in phases}
    for i, (train_idx, valid_idx) in enumerate(cv):
        data = {
            "train": {
                "X": X.iloc[train_idx],
                "a": a.iloc[train_idx],
                "y": y.iloc[train_idx],
            },
            "valid": {
                "X": X.iloc[valid_idx],
                "a": a.iloc[valid_idx],
                "y": y.iloc[valid_idx],
            },
        }
        # TODO: use dict-comprehension to map between phases[0] to cv[0]
        # instead of writing "train" explicitly

        for phase in phases:
            X_fold, a_fold, y_fold = (
                data[phase]["X"],
                data[phase]["a"],
                data[phase]["y"],
            )
            prediction = predictions[phase][i]

            fold_scores = score_estimation(
                prediction, X_fold, a_fold, y_fold, metrics_to_evaluate
            )
            scores[phase].append(fold_scores)

    if isinstance(fold_scores, WeightEvaluatorScores):
        return _combine_weight_evaluator_fold_scores(scores)
    return _combine_fold_scores(scores)

def _get_default_metrics_to_evaluate(first_prediction):
    if isinstance(first_prediction, OutcomePredictions) and not first_prediction.is_binary_outcome:
        metrics_to_evaluate = get_default_regression_metrics()
    else:
        metrics_to_evaluate = get_default_binary_metrics()
    return metrics_to_evaluate


def score_estimation(prediction, X, a_true, y_true, metrics_to_evaluate=None):
    """Should know how to handle the _estimator_predict output provided in `prediction`.
    Can utilize any of the true values provided:
        covariates `X`, treatment assignment `a` or outcome `y`.
    """

    if isinstance(prediction, OutcomePredictions):
        return prediction.evaluate_metrics(a_true, y_true, metrics_to_evaluate)

    # propensity and weight both have the same interface
    # no need to differentiate
    if isinstance(prediction, (PropensityPredictions, WeightPredictions)):
        return prediction.evaluate_metrics(X, a_true, metrics_to_evaluate)
    raise ValueError(f"Invalid type for prediciton: {type(prediction)}")


def _combine_fold_scores(scores):
    """
    Combines scores of each phase and fold into a single object (DataFrame) of scores.

    Args:
        scores (dict[str, list[pd.DataFrame]]): scores of each fold of each phase.
            The structure is {phase_name: [fold_1_score, fold_2_score...]}.
            Where phase_name is usually "train" or "valid", and each fold_i_score
            is a DataFrame which columns are evaluation metrics and rows are
            results of that metrics in that fold.

    Returns:
        pd.DataFrame: Row-concatenated DataFrame with MultiIndex accounting for the
            concatenated folds and phases.
    """
    # Concatenate the scores from list of folds to DataFrame with rows as folds,
    # keeping it by different phases:
    scores = {
        phase: pd.concat(
            scores_fold, axis="index", keys=range(len(scores_fold)), names=["fold"]
        )
        for phase, scores_fold in scores.items()
    }
    # Concatenate the train/validation DataFrame scores into DataFrame with rows as phases:
    scores = pd.concat(scores, axis="index", names=["phase"])
    return scores


def _combine_weight_evaluator_fold_scores(scores):
    # `scores` are provided as WeightEvaluatorScores object for each fold in each phase,
    # Namely, dict[list[WeightEvaluatorScores]], which in turn hold two DataFrames components.
    # In order to combine the underlying DataFrames into a multilevel DataFrame,
    # one must first extract them from the WeightEvaluatorScores object, then recombine.

    # Extract the two components of WeightEvaluatorScores:
    prediction_scores_unfolded = {
        phase: [fold_score.prediction_scores for fold_score in phase_scores]
        for phase, phase_scores in scores.items()
    }
    prediction_scores = _combine_fold_scores(prediction_scores_unfolded)

    covariate_balance_unfolded = {
        phase: [fold_score.covariate_balance for fold_score in phase_scores]
        for phase, phase_scores in scores.items()
    }
    covariate_balance = _combine_fold_scores(covariate_balance_unfolded)

    # Combine the dict[list[DataFrames]] of each component into a multilevel DataFrame separately:
    # TODO: consider reordering the levels, such that the covariate will be the first one and then
    # phase and fold
    # covariate_balance = covariate_balance.reorder_levels(["covariate", "phase", "fold"])

    # Create a new WeightEvaluatorScores object
    # with the combined (i.e., multilevel DataFrame) results:
    scores = WeightEvaluatorScores(prediction_scores, covariate_balance)
    return scores



"""
(C) IBM Corp, 2019, All rights reserved
Created on Aug 25, 2019

@author: EHUD KARAVANI
"""

import unittest
import pandas as pd
import numpy as np
from causallib.estimation import MarginalOutcomeEstimator


class TestMarginalOutcomeEstimator(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.X = pd.DataFrame([[1, 1, 0, 0, 1, 0, 0, 0, 1, 1]])
        cls.y = pd.Series([0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
        cls.a = pd.Series([0, 0, 0, 0, 1, 0, 1, 1, 1, 1])

    def setUp(self):
        self.model = MarginalOutcomeEstimator(learner=None)

    def test_fit_return(self):
        model = self.model.fit(self.X, self.a, self.y)
        self.assertTrue(isinstance(model, MarginalOutcomeEstimator))

    def test_outcome_estimation(self):
        self.model.fit(self.X, self.a, self.y)
        outcomes = self.model.estimate_population_outcome(self.X, self.a, self.y)
        truth = pd.Series([1 / 5, 4 / 5], index=[0, 1])
        pd.testing.assert_series_equal(truth, outcomes)

        with self.subTest("Change covariate and see no change in estimation"):
            X = pd.DataFrame(np.arange(20).reshape(4, 5))  # Different values and shape
            outcomes = self.model.estimate_population_outcome(X, self.a, self.y)
            truth = pd.Series([1/5, 4/5], index=[0, 1])
            pd.testing.assert_series_equal(truth, outcomes)






import abc
import unittest

import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression, LinearRegression

from causallib.estimation import IPW, StratifiedStandardization

from causallib.metrics import get_scorer, get_scorer_names


class BaseTestScorer(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        np.random.seed(0)
        # Data:
        # X, a = make_classification(
        #     n_features=1, n_informative=1, n_redundant=0, n_repeated=0,
        #     n_classes=2, n_clusters_per_class=1, flip_y=0.0, class_sep=0.5,
        #     random_state=0,
        # )
        n = 1000
        d = 2
        X = np.random.normal(scale=3, size=(n, d))
        beta_Xa = np.random.normal(size=d)
        a_logit = X @ beta_Xa  # + np.random.normal(size=n)
        a_propensity = 1 / (1 + np.exp(-a_logit))
        a = np.random.binomial(1, a_propensity)
        beta_Xy = np.random.normal(size=d)
        beta_a = 5
        y = X @ beta_Xy + a * beta_a + np.random.normal(size=n)
        cls.data = {
            "X": pd.DataFrame(X),
            "a": pd.Series(a),
            "y": pd.Series(y),
        }

        # RCT Data:
        np.random.seed(0)
        X = np.random.normal(0, 1, 1000)
        a = np.random.binomial(1, 0.5, 1000)
        y = X * 1 + a * 5 + np.random.normal(size=X.shape[0])
        cls.data_rct = {
            "X": pd.DataFrame(X),
            "a": pd.Series(a),
            "y": pd.Series(y),
        }

        # # Avoids regularization of the model:
        cls.estimator = IPW(LogisticRegression(penalty='none', solver='sag', max_iter=2000))

    def ensure_single_scoring_does_not_fail(self, scoring_name, data):
        scorer = get_scorer(scoring_name)
        score = scorer(
            self.estimator,
            data['X'], data['a'], data['y'],
        )
        self.assertIsInstance(score, float)
        self.assertLess(score, 0)

    def ensure_multiple_scoring_do_not_fail(self, score_type):
        self.estimator.fit(self.data['X'], self.data['a'], self.data['y'])
        scoring_names = get_scorer_names(score_type)
        for scoring_name in scoring_names:
            with self.subTest(f"Testing {scoring_name} does not fail."):
                self.ensure_single_scoring_does_not_fail(scoring_name, self.data)

    @abc.abstractmethod
    @unittest.skip
    def test_multiple_scoring_do_not_fail(self):
        pass


class TestPropensityScorer(BaseTestScorer):

    def test_multiple_scoring_do_not_fail(self):
        self.ensure_multiple_scoring_do_not_fail("propensity")

    def test_scoring_with_kwargs(self):
        self.estimator.fit(self.data['X'], self.data['a'], self.data['y'])

        with self.subTest("single known kwarg:"):
            scorer = get_scorer("weighted_roc_curve_error")
            score_default = scorer(
                self.estimator,
                self.data['X'], self.data['a'], self.data['y'],
            )
            score_agg_min = scorer(
                self.estimator,
                self.data['X'], self.data['a'], self.data['y'],
                agg=np.min,
            )
            self.assertLess(abs(score_agg_min), abs(score_default))
            # Default is max. scores are negative metrics values due to lower-is-better

        with self.subTest("multiple known kwargs"):
            scorer = get_scorer("ici_error")
            score_default = scorer(
                self.estimator,
                self.data['X'], self.data['a'], self.data['y'],
            )
            score_kwargs = scorer(
                self.estimator,
                self.data['X'], self.data['a'], self.data['y'],
                agg=np.max,
                lowess_kwargs=dict(frac=0.2),
            )
            self.assertLess(abs(score_default), abs(score_kwargs))
            # Default is mean. scores are negative metrics values due to lower-is-better

        # @unittest.expectedFailure
        # with self.subTest("passing unknown kwargs"):
        #     propensity_scoring_names = get_scorer_names("propensity")
        #     for scoring_name in propensity_scoring_names:
        #         with self.subTest(f"Testing unknown kwarg in {scoring_name}."):
        #             scorer = get_scorer(scoring_name)
        #             score = scorer(
        #                 self.estimator,
        #                 self.data['X'], self.data['a'], self.data['y'],
        #                 nonexistingkwarg="shouldn't_exist!",
        #             )
        #             self.assertLess(0, score)


class TestWeightScorer(BaseTestScorer):
    def test_multiple_scoring_do_not_fail(self):
        self.ensure_multiple_scoring_do_not_fail("weight")

    def test_scoring_with_kwargs(self):
        self.estimator.fit(self.data['X'], self.data['a'], self.data['y'])
        scorer = get_scorer("covariate_balancing_error")
        score_default = scorer(
            self.estimator,
            self.data['X'], self.data['a'], self.data['y'],
        )
        score_agg_min = scorer(
            self.estimator,
            self.data['X'], self.data['a'], self.data['y'],
            agg=np.min,
        )
        self.assertLess(-score_agg_min, -score_default)  # Default is max. Scores are negative metrics values


class TestOutcomeScorer(BaseTestScorer):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.estimator = StratifiedStandardization(LinearRegression())

    def test_multiple_scoring_do_not_fail(self):
        self.ensure_multiple_scoring_do_not_fail("outcome")

    def test_scoring_with_kwargs(self):
        self.estimator.fit(self.data['X'], self.data['a'], self.data['y'])
        scorer = get_scorer("balanced_residuals_error")
        score_default = scorer(
            self.estimator,
            self.data['X'], self.data['a'], self.data['y'],
        )
        score_mean_diff = scorer(
            self.estimator,
            self.data['X'], self.data['a'], self.data['y'],
            distance_metric=lambda y0, y1: np.abs(y1.mean() - y0.mean())
        )
        self.assertLess(-score_default, -score_mean_diff)  # Default is standardized, scores are negatives

    def test_multiple_outcome_models(self):
        from causallib.estimation import (
            Standardization, StratifiedStandardization,
            PropensityFeatureStandardization,
            WeightedStandardization,
            TMLE,
            RLearner,
        )
        models = [
            Standardization(LinearRegression()),
            StratifiedStandardization(LinearRegression()),
            PropensityFeatureStandardization(
                Standardization(LinearRegression()),
                IPW(LogisticRegression())
            ),
            WeightedStandardization(
                Standardization(LinearRegression()),
                IPW(LogisticRegression())
            ),
            TMLE(
                Standardization(LinearRegression()),
                IPW(LogisticRegression())
            ),
            RLearner(
                LinearRegression(),
                LinearRegression(),
                LogisticRegression(),
            )
        ]
        for model in models:
            with self.subTest(f"Test scoring of model"):
                model.fit(self.data['X'], self.data['a'], self.data['y'])
                scorer = get_scorer("balanced_residuals_error")
                score = scorer(
                    model,
                    self.data['X'], self.data['a'], self.data['y'],
                )
                self.assertIsInstance(score, float)




import unittest
import pandas as pd
import numpy as np

from sklearn.linear_model import LogisticRegression, LinearRegression
from sklearn.ensemble import GradientBoostingRegressor, GradientBoostingClassifier
from sklearn.model_selection import GridSearchCV

from causallib.estimation import IPW
from causallib.estimation import AIPW
from causallib.estimation import StratifiedStandardization, Standardization
from causallib.metrics import get_scorer

from causallib.model_selection import causalize_searcher


class TestGridSearch(unittest.TestCase):
    @classmethod
    def _generate_data(cls, n=100, d=4):
        np.random.seed(1)
        priors = np.random.uniform(0.1, 0.9, size=d)
        X = np.random.binomial(1, priors, size=(n, d))

        beta_a = np.random.normal(size=d)
        intercept_a = np.random.normal()
        a_logit = intercept_a + X @ beta_a + np.random.normal(size=n)
        a_propensity = 1 / (1 + np.exp(-a_logit))
        a = np.random.binomial(1, a_propensity)

        beta_y = np.random.normal(size=d)
        intercept_y = np.random.normal()
        treatment_effect = -3
        y = intercept_y + X @ beta_y + treatment_effect * a + np.random.normal(size=n)

        X = pd.DataFrame(X, columns=[f"x{i}" for i in range(X.shape[1])])
        a = pd.Series(a, name="treatment")
        y = pd.Series(y, name="outcome")
        data = dict(X=X, a=a, y=y)
        return data

    @classmethod
    def setUpClass(cls):
        cls.data = cls._generate_data()

    def _fit_search_model(self, estimator, scoring, param_grid, data=None, cv=2, **search_kwargs):
        if data is None:
            data = self.data
        CausalGridSearchCV = causalize_searcher(GridSearchCV)
        grid_model = CausalGridSearchCV(
            estimator,
            param_grid=param_grid,
            scoring=scoring,
            cv=cv,
            **search_kwargs,
        )
        grid_model.fit(data['X'], data['a'], data['y'])
        return grid_model

    def test_single_propensity_scorer(self):
        data = self.data
        model = IPW(LogisticRegression(penalty="none", solver="saga", max_iter=10000))
        scorer = get_scorer("weighted_roc_auc_error")
        grid_model = self._fit_search_model(model, scorer, dict(clip_min=[0.2, 0.3]))

        self.assertIsInstance(grid_model.best_estimator_, model.__class__)
        grid_model.best_estimator_.estimate_population_outcome(data['X'], data['a'], data['y'])
        pd.testing.assert_series_equal(
            grid_model.best_estimator_.estimate_population_outcome(data['X'], data['a'], data['y']),
            grid_model.estimate_population_outcome(data['X'], data['a'], data['y'])
        )

        with self.subTest("Grid model has its inner estimator's methods that work as expected"):
            propensities = grid_model.compute_propensity(data['X'], data['a'])
            self.assertLessEqual(0.2, propensities.min())

            weights = grid_model.compute_weights(data['X'], data['a'])
            self.assertIsInstance(weights, pd.Series)
            weights = grid_model.compute_weight_matrix(data['X'], data['a'])
            self.assertIsInstance(weights, pd.DataFrame)

    def test_single_outcome_scorer(self):
        data = self.data
        model = StratifiedStandardization(GradientBoostingRegressor())
        # Can handle the `learner` being a dict because before `fit` it is not a dict, but a single model type
        scorer = get_scorer("balanced_residuals_error")
        grid_model = self._fit_search_model(model, scorer, dict(learner__n_estimators=[10, 20]))

        self.assertIsInstance(grid_model.best_estimator_, model.__class__)
        pd.testing.assert_series_equal(
            grid_model.best_estimator_.estimate_population_outcome(data['X'], data['a'], data['y']),
            grid_model.estimate_population_outcome(data['X'], data['a'], data['y'])
        )

        with self.subTest("Estimator behaves as inner estimator"):
            ind_outcomes = grid_model.estimate_individual_outcome(data['X'], data['a'])
            self.assertEqual(data['X'].shape[0], ind_outcomes.shape[0])
            self.assertEqual(2, ind_outcomes.shape[1])

    # @unittest.expectedFailure
    def test_different_stratified_estimators(self):
        """Should fail because can't assign the params once the `learner` is a dictionary."""
        data = self.data
        model = StratifiedStandardization({
            0: GradientBoostingRegressor(), 1: LinearRegression(),
        })
        scorer = get_scorer("balanced_residuals_error")
        with self.assertRaises(AttributeError):
            grid_model = self._fit_search_model(model, scorer, dict(learner__n_estimators=[10, 20]))
        # self.assertIsInstance(grid_model.best_estimator_, model.__class__)
        # grid_model.best_estimator_.estimate_population_outcome(data['X'], data['a'], data['y'])

    def test_single_propensity_scorer_name(self):
        data = self.data
        model = IPW(LogisticRegression(penalty="none", solver="saga", max_iter=10000))
        grid_model = self._fit_search_model(model, "weighted_roc_auc_error", dict(clip_min=[0.2, 0.3]))

        self.assertIsInstance(grid_model.best_estimator_, model.__class__)
        grid_model.best_estimator_.estimate_population_outcome(data['X'], data['a'], data['y'])
        pd.testing.assert_series_equal(
            grid_model.best_estimator_.estimate_population_outcome(data['X'], data['a'], data['y']),
            grid_model.estimate_population_outcome(data['X'], data['a'], data['y'])
        )

    def test_multiple_scorers(self):
        data = self.data
        model = IPW(LogisticRegression(penalty="none", solver="saga", max_iter=10000, random_state=0))
        scorers_dict = {
            "weighted_roc_auc_error": get_scorer("weighted_roc_auc_error"),
            "covariate_balancing_error": get_scorer("covariate_balancing_error"),
        }
        scorers_list = ["weighted_roc_auc_error", "covariate_balancing_error"]
        scorers_tuple = ("weighted_roc_auc_error", "covariate_balancing_error")
        scorers_set = {"weighted_roc_auc_error", "covariate_balancing_error"}
        grid_model_dict = self._fit_search_model(
            model, scorers_dict,
            dict(clip_min=[0.05, 0.2]),
            refit="weighted_roc_auc_error",
        )
        grid_model_list = self._fit_search_model(
            model, scorers_list,
            dict(clip_min=[0.05, 0.2]),
            refit="weighted_roc_auc_error"
        )
        grid_model_tuple = self._fit_search_model(
            model, scorers_tuple,
            dict(clip_min=[0.05, 0.2]),
            refit="weighted_roc_auc_error"
        )
        grid_model_set = self._fit_search_model(
            model, scorers_set,
            dict(clip_min=[0.05, 0.2]),
            refit="weighted_roc_auc_error"
        )

        # dict with list
        self.assertEqual(
            grid_model_dict.best_estimator_.clip_min,
            grid_model_list.best_estimator_.clip_min,
        )
        np.testing.assert_array_almost_equal(
            grid_model_dict.best_estimator_.learner.coef_,
            grid_model_list.best_estimator_.learner.coef_,
        )
        pd.testing.assert_series_equal(
            grid_model_dict.estimate_population_outcome(data['X'], data['a'], data['y']),
            grid_model_list.estimate_population_outcome(data['X'], data['a'], data['y']),
            check_exact=False,
        )
        # tuple with list
        self.assertEqual(
            grid_model_tuple.best_estimator_.clip_min,
            grid_model_list.best_estimator_.clip_min,
        )
        np.testing.assert_array_almost_equal(
            grid_model_tuple.best_estimator_.learner.coef_,
            grid_model_list.best_estimator_.learner.coef_,
        )
        pd.testing.assert_series_equal(
            grid_model_tuple.estimate_population_outcome(data['X'], data['a'], data['y']),
            grid_model_list.estimate_population_outcome(data['X'], data['a'], data['y']),
            check_exact=False,
        )
        # set with list
        self.assertEqual(
            grid_model_set.best_estimator_.clip_min,
            grid_model_list.best_estimator_.clip_min,
        )
        np.testing.assert_array_almost_equal(
            grid_model_set.best_estimator_.learner.coef_,
            grid_model_list.best_estimator_.learner.coef_,
        )
        pd.testing.assert_series_equal(
            grid_model_set.estimate_population_outcome(data['X'], data['a'], data['y']),
            grid_model_list.estimate_population_outcome(data['X'], data['a'], data['y']),
            check_exact=False,
        )

    def test_bad_scorer(self):
        model = IPW(LogisticRegression(penalty="none", solver="saga", max_iter=10000))
        CausalGridSearchCV = causalize_searcher(GridSearchCV)
        with self.assertRaises(ValueError):
            grid_model = CausalGridSearchCV(
                model,
                param_grid=dict(clip_min=[0.2, 0.3]),
                scoring=3,
            )

    def test_doubly_robust(self):
        data = self.data
        CausalGridSearchCV = causalize_searcher(GridSearchCV)

        ipw = IPW(LogisticRegression(penalty="none", solver="saga", max_iter=10000))
        ipw_grid_model = CausalGridSearchCV(
            ipw,
            param_grid=dict(clip_min=[0.2, 0.3]), cv=2,
            scoring=get_scorer("weighted_roc_auc_error"),
        )

        std = Standardization(GradientBoostingRegressor())
        CausalGridSearchCV = causalize_searcher(GridSearchCV)
        std_grid_model = CausalGridSearchCV(
            std,
            param_grid=dict(learner__n_estimators=[10, 20]), cv=3,
            scoring=get_scorer("balanced_residuals_error"),
        )

        model = AIPW(
            outcome_model=std_grid_model,
            weight_model=ipw_grid_model,
        )

        model.fit(data['X'], data['a'], data['y'])
        propensities = model.weight_model.compute_propensity(data['X'], data['a'])
        self.assertLessEqual(0.2, propensities.min())
        model.estimate_population_outcome(data['X'], data['a'], data['y'])
        pd.testing.assert_frame_equal(
            model.estimate_individual_outcome(data['X'], data['a']),
            model.outcome_model.estimate_individual_outcome(data['X'], data['a'])
            # AIPW's individual outcome is the internal standardization individual outcome
        )

    def test_with_survival_ipw(self):
        from causallib.survival import WeightedSurvival, WeightedStandardizedSurvival
        from causallib.datasets import load_nhefs_survival
        from sklearn.dummy import DummyClassifier

        data = load_nhefs_survival()
        idx = data.X.sample(n=100, random_state=0).index
        X, a, t, y = data.X.loc[idx], data.a.loc[idx], data.t.loc[idx], data.y.loc[idx]
        CausalGridSearchCV = causalize_searcher(GridSearchCV)

        ipw = IPW(LogisticRegression(penalty="none", solver="saga", max_iter=10000))
        ipw_grid_model = CausalGridSearchCV(
            ipw,
            param_grid=dict(clip_min=[0.2, 0.3]), cv=2,
            scoring="weighted_roc_auc_error",
        )

        for Model in [WeightedSurvival, WeightedStandardizedSurvival]:
            model = Model(
                weight_model=ipw_grid_model,
                survival_model=DummyClassifier(strategy="prior")
            )
            model.fit(X, a, t, y)

            self.assertIsInstance(model.weight_model, CausalGridSearchCV)

            propensities = model.weight_model.compute_propensity(X, a)
            self.assertLessEqual(0.2, propensities.min())

            outcomes = model.estimate_population_outcome(X, a, t, y, timeline_start=1)
            self.assertEqual(outcomes.shape, (t.max(), a.nunique()))

    def test_selecting_different_core_estimators(self):
        from sklearn.dummy import DummyClassifier
        data = self.data
        model = IPW(DummyClassifier())  # DummyClassifier used just for initialization
        scorer = get_scorer("weighted_roc_auc_error")
        param_grid = dict(
            learner=[
                LogisticRegression(penalty="none", solver="saga", max_iter=10000, random_state=0),
                LogisticRegression(penalty="l1", solver="saga", max_iter=10000, random_state=0),
                LogisticRegression(penalty="l2", solver="saga", max_iter=10000, random_state=0),
                GradientBoostingClassifier(random_state=0),
            ]
        )
        grid_model = self._fit_search_model(
            model,
            scorer,
            param_grid
        )

        self.assertIsInstance(grid_model.best_estimator_, model.__class__)
        self.assertNotIsInstance(grid_model.best_estimator_, DummyClassifier)
        self.assertIsInstance(grid_model.best_estimator_.learner, LogisticRegression)  # Best base estimator
        outcomes = grid_model.best_estimator_.estimate_population_outcome(data['X'], data['a'], data['y'])
        self.assertIsInstance(outcomes, pd.Series)

    def test_different_search_algorithm(self):
        from causallib.model_selection import RandomizedSearchCV
        from scipy.stats import uniform

        data = self.data
        ipw = IPW(LogisticRegression(penalty="none", solver="saga", max_iter=10000))
        model = RandomizedSearchCV(
            ipw,
            param_distributions=dict(clip_min=uniform(loc=0, scale=0.5)), cv=2,
            scoring=get_scorer("weighted_roc_auc_error"),
        )
        model.fit(data['X'], data['a'], data['y'])
        self.assertIsInstance(model.best_estimator_, ipw.__class__)
        model.best_estimator_.estimate_population_outcome(data['X'], data['a'], data['y'])
        pd.testing.assert_series_equal(
            model.best_estimator_.estimate_population_outcome(data['X'], data['a'], data['y']),
            model.estimate_population_outcome(data['X'], data['a'], data['y'])
        )

    def test_with_custom_kfold(self):
        from causallib.model_selection import TreatmentOutcomeStratifiedKFold
        from causallib.model_selection import TreatmentStratifiedKFold

        data = self.data
        y = np.random.binomial(2, 0.5, size=data['y'].shape[0])  # binary y, for doubly stratification
        y = pd.Series(y, index=data['y'].index)
        model = IPW(LogisticRegression(penalty="none", solver="saga", max_iter=10000))
        scorer = get_scorer("weighted_roc_auc_error")

        with self.subTest("Doubly Stratified K-fold"):
            cv = TreatmentOutcomeStratifiedKFold(n_splits=2)
            treatment_outcome_grid_model = self._fit_search_model(
                model, scorer,
                dict(clip_min=[0.2, 0.3]),
                cv=cv,
                data={**data, "y": y},
            )
            outcomes = treatment_outcome_grid_model.estimate_population_outcome(data['X'], data['a'], data['y'])
            self.assertIsInstance(outcomes, pd.Series)

        with self.subTest("Treatment-stratified k-fold"):
            cv = TreatmentStratifiedKFold(n_splits=2)
            treatment_grid_model = self._fit_search_model(
                model, scorer,
                dict(clip_min=[0.2, 0.3]),
                cv=cv,
                data={**data, "y": y},
            )
            outcomes = treatment_grid_model.estimate_population_outcome(data['X'], data['a'], data['y'])
            self.assertIsInstance(outcomes, pd.Series)



"""Mixins for plotting.

To work the mixin requires the class to implement `get_data_for_plot` with the
supported plot names. See .data_extractors for examples. """

from . import plots


class WeightPlotterMixin:
    """Mixin to add members to for weight estimation plotting.

    Class must implement:
      * `get_data_for_plot(plots.COVARIATE_BALANCE_GENERIC_PLOT)`
      * `get_data_for_plot(plots.WEIGHT_DISTRIBUTION_PLOT)`
    """

    def plot_covariate_balance(
        self,
        kind="love",
        phase="train",
        ax=None,
        aggregate_folds=True,
        thresh=None,
        plot_semi_grid=True,
        **kwargs,
    ):
        """Plot covariate balance before and after weighting.

        Args:
            kind (str, optional): Plot kind, "love" ,"slope" or "scatter". Defaults to "love".
            phase (str, optional): Phase to plot: "train" or "valid". Defaults to "train".
            ax (matplotlib.axes.Axes, optional): axis to plot on, if None creates new axis.
                Defaults to None.
            aggregate_folds (bool, optional): Whether to aggregate folds. Defaults to True.
                Ignored when kind="slope".
            thresh (float, optional): Draw threshold line at value. Defaults to None.
            plot_semi_grid (bool, optional): Defaults to True. Ignored when kind="slope".

        Returns:
            matplotlib.axes.Axes: axis with plot
        """
        (table1_folds,) = self.get_data_for_plot(
            plots.COVARIATE_BALANCE_GENERIC_PLOT, phase=phase
        )
        if kind == "love":
            return plots.plot_mean_features_imbalance_love_folds(
                table1_folds=table1_folds,
                ax=ax,
                aggregate_folds=aggregate_folds,
                thresh=thresh,
                plot_semi_grid=plot_semi_grid,
                **kwargs,
            )
        if kind == "slope":
            return plots.plot_mean_features_imbalance_slope_folds(
                table1_folds=table1_folds,
                ax=ax,
                thresh=thresh,
                **kwargs,
            )
    

        if kind == "scatter":
            return plots.plot_mean_features_imbalance_scatter_plot(
                table1_folds=table1_folds,
                ax=ax,
                thresh=thresh,
                **kwargs,
            )
        


        raise ValueError(f"Unsupported covariate balance plot kind {kind}")

    def plot_weight_distribution(
        self,
        phase="train",
        reflect=True,
        kde=False,
        cumulative=False,
        norm_hist=True,
        ax=None,
    ):
        """
        Plot the distribution of propensity score.

        Args:
            phase (str, optional): Phase to plot: "train" or "valid". Defaults to "train".
            reflect (bool): Whether to plot treatment groups on opposite sides of the x-axis.
                This can only work if there are exactly two groups.
            kde (bool): Whether to plot kernel density estimation
            cumulative (bool): Whether to plot cumulative distribution.
            norm_hist (bool): If False - use raw counts on the y-axis.
                            If kde=True, then norm_hist should be True as well.
            ax (matplotlib.axes.Axes, optional): axis to plot on, if None creates new axis.
                Defaults to None.
        Returns:
            matplotlib.axes.Axes

        """
        weights, treatments, cv = self.get_data_for_plot(
            plots.WEIGHT_DISTRIBUTION_PLOT, phase=phase
        )

        return plots.plot_propensity_score_distribution_folds(
            predictions=weights,
            hue_by=treatments,
            cv=cv,
            reflect=reflect,
            kde=kde,
            cumulative=cumulative,
            norm_hist=norm_hist,
            ax=ax,
        )


class ClassificationPlotterMixin:
    """Mixin to add members to for classification/binary prediction estimation.

    This occurs for propensity models (treatment assignment is inherently binary)
    and for outcome models where the outcome is binary.

    Class must implement:
      * `get_data_for_plot(plots.ROC_CURVE_PLOT)`
      * `get_data_for_plot(plots.PR_CURVE_PLOT)`
      * `get_data_for_plot(plots.CALIBRATION_PLOT)`
    """

    def plot_roc_curve(
        self,
        phase="train",
        plot_folds=False,
        label_folds=False,
        label_std=False,
        ax=None,
    ):
        """Plot ROC curve.

        Args:
            phase (str, optional): Phase to plot: "train" or "valid". Defaults to "train".
            plot_folds (bool, optional): Whether to plot individual folds. Defaults to False.
            label_folds (bool, optional): Whether to label folds. Defaults to False.
            label_std (bool, optional): Whether to label std. Defaults to False.
            ax (matplotlib.axes.Axes, optional): axis to plot on, if None creates new axis.
                Defaults to None.
        Returns:
            matplotlib.axes.Axes
        """
        (roc_curve_data,) = self.get_data_for_plot(plots.ROC_CURVE_PLOT, phase=phase)
        return plots.plot_roc_curve_folds(
            roc_curve_data,
            ax=ax,
            plot_folds=plot_folds,
            label_folds=label_folds,
            label_std=label_std,
        )

    def plot_pr_curve(
        self,
        phase="train",
        plot_folds=False,
        label_folds=False,
        label_std=False,
        ax=None,
    ):
        """Plot precision-recall (PR) curve.

        Args:
            phase (str, optional): Phase to plot: "train" or "valid". Defaults to "train".
            plot_folds (bool, optional): Whether to plot individual folds. Defaults to False.
            label_folds (bool, optional): Whether to label folds. Defaults to False.
            label_std (bool, optional): Whether to label std. Defaults to False.
            ax (matplotlib.axes.Axes, optional): axis to plot on, if None creates new axis.
                Defaults to None.
        Returns:
            matplotlib.axes.Axes
        """
        (pr_curve_data,) = self.get_data_for_plot(plots.PR_CURVE_PLOT, phase=phase)
        return plots.plot_precision_recall_curve_folds(
            pr_curve_data,
            ax=ax,
            plot_folds=plot_folds,
            label_folds=label_folds,
            label_std=label_std,
        )

    def plot_calibration_curve(
        self,
        phase="train",
        n_bins=10,
        plot_se=True,
        plot_rug=False,
        plot_histogram=False,
        quantile=False,
        ax=None,
    ):
        """Plot calibration curves for multiple models (presumably in folds)

        Args:
            phase (str, optional): Phase to plot: "train" or "valid". Defaults to "train".
            n_bins (int): number of bins to evaluate in the plot
            plot_se (bool): Whether to plot standard errors around the mean
                bin-probability estimation.
            plot_rug (bool):
            plot_histogram (bool):
            quantile (bool): If true, the binning of the calibration curve is by quantiles.
                Defaults to False.
            ax (matplotlib.axes.Axes, optional): axis to plot on, if None creates new axis.
                Defaults to None.
        Note:
            One of plot_propensity or plot_model must be True.

        Returns:
            matplotlib.axes.Axes
        """

        predictions, targets, cv = self.get_data_for_plot(
            plots.CALIBRATION_PLOT, phase=phase
        )
        return plots.plot_calibration_folds(
            predictions=predictions,
            targets=targets,
            cv=cv,
            n_bins=n_bins,
            plot_se=plot_se,
            plot_rug=plot_rug,
            plot_histogram=plot_histogram,
            quantile=quantile,
            ax=ax,
        )


class ContinuousOutcomePlotterMixin:
    """Mixin to add members to for continous outcome estimation.

    Class must implement:
      * `get_data_for_plot(plots.CONTINUOUS_ACCURACY_PLOT)`
      * `get_data_for_plot(plots.RESIDUALS_PLOT)`
      * `get_data_for_plot(plots.CONTINUOUS_ACCURACY_PLOT)`
    """

    def plot_continuous_accuracy(
        self, phase="train", alpha_by_density=True, plot_residuals=False, ax=None
    ):
        """Plot continuous accuracy,

        Args:
            phase (str, optional): Phase to plot: "train" or "valid". Defaults to "train".
            alpha_by_density (bool, optional): Whether to calculate points alpha value
                (transparent-opaque) with density estimation. This can take some time
                to compute for a large number of points. If False, alpha calculation
                will be a simple fast heuristic.
            plot_residuals (bool, optional): Whether to plot residuals. Defaults to False.
            ax (matplotlib.axes.Axes, optional): axis to plot on, if None creates new axis.
                Defaults to None.
        Returns:
            matplotlib.axes.Axes
        """
        predictions, y, a, cv = self.get_data_for_plot(
            plots.CONTINUOUS_ACCURACY_PLOT, phase=phase
        )
        return plots.plot_continuous_prediction_accuracy_folds(
            predictions=predictions,
            y=y,
            a=a,
            cv=cv,
            alpha_by_density=alpha_by_density,
            plot_residuals=plot_residuals,
            ax=ax,
        )

    def plot_residuals(self, phase="train", alpha_by_density=True, ax=None):
        """Plot residuals of predicted outcome vs ground truth.

        Args:
            phase (str, optional): Phase to plot: "train" or "valid". Defaults to "train".
            alpha_by_density (bool, optional): Whether to calculate points alpha value
                (transparent-opaque) with density estimation. This can take some time
                to compute for a large number of points. If False, alpha calculation
                will be a simple fast heuristic.
            ax (matplotlib.axes.Axes, optional): axis to plot on, if None creates new axis.
                Defaults to None.
        Returns:
            matplotlib.axes.Axes
        """
        predictions, y, a, cv = self.get_data_for_plot(
            plots.RESIDUALS_PLOT, phase=phase
        )

        return plots.plot_residual_folds(
            predictions=predictions,
            y=y,
            a=a,
            cv=cv,
            alpha_by_density=alpha_by_density,
            ax=ax,
        )

    def plot_common_support(self, phase="train", alpha_by_density=True, ax=None):
        """Plot the scatter plot of y0 vs. y1 for multiple scoring results, colored by the treatment

        Args:
            alpha_by_density (bool): Whether to calculate points alpha value (transparent-opaque)
               with density estimation. This can take some time to compute for a large number
               of points. If False, alpha calculation will be a simple fast heuristic.
            ax (plt.Axes): The axes on which the plot will be displayed. Optional.
        """
        predictions, treatments, cv = self.get_data_for_plot(
            plots.COMMON_SUPPORT_PLOT, phase=phase
        )
        return plots.plot_counterfactual_common_support_folds(
            predictions=predictions,
            hue_by=treatments,
            cv=cv,
            alpha_by_density=alpha_by_density,
            ax=ax,
        )


class PlotAllMixin:
    """Mixin to make all the train and validation plots.

    Class must implement:
      * `all_plot_names`
      * `get_data_for_plot(name)` for every name in `all_plot_names`
    """

    def plot_all(self, phase=None):
        """Create plot of all available EvaluationResults.

        Will create a figure with a subplot for each plot name in `all_plot_names`.
        If `results` have train and validation data, will create separate
        "train" and "valid" figures. If a single plot is requested, only that plot is created.

        Args:
            phase (Union[str, None], optional): phase to plot "train" or "valid". If not supplied,
                defaults to both if available.

        Returns:
            Dict[str, matplotlib.axis.Axis]]: the Axis objects of the plots in a nested dictionary:
              * First key is the phase ("train" or "valid")
              * Second key is the plot name.
        """
        phases_to_plot = self.predictions.keys() if phase is None else [phase]
        multipanel_plot = {
            plotted_phase: self._make_multipanel_evaluation_plot(
                plot_names=self.all_plot_names, phase=plotted_phase
            )
            for plotted_phase in phases_to_plot
        }
        return multipanel_plot

    def _make_multipanel_evaluation_plot(self, plot_names, phase):
        phase_fig, phase_axes = plots.get_subplots(len(plot_names))
        named_axes = {
            name: self._make_single_panel_evaluation_plot(name, phase, ax)
            for name, ax in zip(plot_names, phase_axes.ravel())
        }

        phase_fig.suptitle(f"Evaluation on {phase} phase")
        return named_axes

    def _make_single_panel_evaluation_plot(self, plot_name, phase, ax=None, **kwargs):
        """Create a single evaluation plot.

        For a single phase and a single plot name.

        Args:
            results (EvaluationResults): evaluation results to plot
            plot_name (str): plot name (from results.all_plot_names)
            phase (str): "train" or "valid"
            ax (matplotlib.axis.Axis, optional): axis to plot on. Defaults to None.
            **kwargs: passed to underlying plotting function

        Returns:
            Union[matplotlib.axis.Axis, None]: axis with plot if successful, else None
        """

        plot_func = plots.lookup_name(plot_name)
        plot_data = self.get_data_for_plot(plot_name, phase=phase)
        return plot_func(*plot_data, ax=ax, **kwargs)



from causallib.estimation.base_weight import WeightEstimator
from .univariate_curve_fitter import UnivariateCurveFitter
from sklearn.base import BaseEstimator as SKLearnBaseEstimator
from typing import Any
import pandas as pd
from copy import deepcopy
from .survival_utils import canonize_dtypes_and_names
from .base_survival import SurvivalBase
from typing import Optional


class WeightedSurvival(SurvivalBase):
    """
    Weighted survival estimator
    """

    def __init__(self,
                 weight_model: WeightEstimator = None,
                 survival_model: Any = None):
        """
        Weighted survival estimator.
        Args:
            weight_model: causallib compatible weight model (e.g., IPW)
            survival_model: Three alternatives:
                1. None - compute non-parametric KaplanMeier survival curve
                2. Scikit-Learn estimator (needs to implement `predict_proba`) - compute parametric curve by fitting a
                    time-varying hazards model
                3. lifelines UnivariateFitter - use lifelines fitter to compute survival curves from events and durations
        """
        self.weight_model = weight_model

        # Construct default curve fitter, non parametric estimation (Kaplan-Meier)
        if survival_model is None:
            self.survival_model = UnivariateCurveFitter()
        # Construct default curve fitter, parametric with a scikit-learn estimator
        elif isinstance(survival_model, SKLearnBaseEstimator):
            self.survival_model = UnivariateCurveFitter(survival_model)
        # Initialized lifelines univariate fitter (or any implementation with a compatible API)
        else:
            self.survival_model = survival_model

    def fit(self,
            X: pd.DataFrame,
            a: pd.Series,
            t: pd.Series = None,
            y: pd.Series = None,
            fit_kwargs: Optional[dict] = None):
        """
        Fits internal weight module (e.g. IPW module, adversarial weighting, etc).

        Args:
            X (pd.DataFrame): Baseline covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            t (pd.Series): NOT USED (for compatibility only)
            y (pd.Series): NOT USED (for compatibility only)
            fit_kwargs (dict): Optional kwargs for fit call of survival model (NOT USED, since fit
                               call of survival model occurs in 'estimate_population_outcome' rather than here)

        Returns:
            self
        """
        a, _, y, _, X = canonize_dtypes_and_names(a=a, t=None, y=y, w=None, X=X)
        if self.weight_model is not None:
            self.weight_model.fit(X=X, a=a, y=y)

        return self

    def estimate_population_outcome(self,
                                    X: pd.DataFrame,
                                    a: pd.Series,
                                    t: pd.Series,
                                    y: pd.Series,
                                    timeline_start: Optional[int] = None,
                                    timeline_end: Optional[int] = None
                                    ) -> pd.DataFrame:
        """
        Returns population averaged survival curves.

        Args:
            X (pd.DataFrame): Baseline covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            t (pd.Series|int): Followup durations, size (num_subjects,).
            y (pd.Series): Observed outcome (1) or right censoring event (0), size (num_subjects,).
            timeline_start (int): Common start time-step. If provided, will generate survival curves starting
                                  from 'timeline_start' for all patients. If None, will predict from first observed event.
            timeline_end (int): Common end time-step. If provided, will generate survival curves up to 'timeline_end'
                                for all patients. If None, will predict up to last observed event.

        Returns:
            pd.DataFrame: with timestep index, treatment values as columns and survival as entries
        """
        self.stratified_curve_fitters_ = {}
        a, t, y, _, X = canonize_dtypes_and_names(a=a, t=t, y=y, w=None, X=X)
        min_time = timeline_start if timeline_start is not None else int(t.min())
        max_time = timeline_end if timeline_end is not None else int(t.max())

        if self.weight_model is not None:
            # Generate inverse propensity for treatment weights (IPTW)
            iptw_weights = self.weight_model.compute_weights(X, a)
            iptw_weights.name = 'w'
        else:
            iptw_weights = None

        # Fit or compute survival curves
        treatment_values = a.unique()
        survival_curves = []
        for treatment_value in treatment_values:
            stratum_indices = a == treatment_value
            stratum_curve_fitter = deepcopy(self.survival_model)

            # Fit curve model
            stratum_curve_fitter.fit(durations=t[stratum_indices], event_observed=y[stratum_indices],
                                     weights=iptw_weights[stratum_indices] if iptw_weights is not None else None)
            self.stratified_curve_fitters_[treatment_value] = stratum_curve_fitter

            # Predict curve model
            curve = stratum_curve_fitter.predict(times=range(min_time, max_time + 1))
            curve.rename(treatment_value, inplace=True)
            survival_curves.append(curve)

        res = pd.concat(survival_curves, axis=1)

        # Setting index/column names
        res.index.name = t.name
        res.columns.name = a.name
        return res



"""Predictor classes.

Predictors generate sets of predictions for a single fold with no cross-validation
or train-test logic.
"""

import abc
from copy import deepcopy
from typing import Union

import pandas as pd

from ..estimation.base_estimator import IndividualOutcomeEstimator
from ..estimation.base_weight import PropensityEstimator, WeightEstimator
from ..utils.stat_utils import robust_lookup

from .predictions import PropensityPredictions, WeightPredictions, OutcomePredictions


def predict_cv(estimator, X, a, y, cv, refit=True, phases=("train", "valid")):
    """Obtain predictions on the provided data in cross-validation

    Args:
        X (pd.DataFrame): Covariates.
        a (pd.Series): Treatment assignment.
        y (pd.Series): Outcome.
        cv (list[tuples]): list the number of folds containing tuples of indices
            (train_idx, validation_idx)
        refit (bool): Whether to refit the model on each fold.
        phases (list[str]): {["train", "valid"], ["train"], ["valid"]}.
            Phases names to evaluate on - train ("train"), validation ("valid") or both.
            'train' corresponds to cv[i][0] and 'valid' to  cv[i][1]
    Returns:
        (dict[str, list], list): A two-tuple containing:

            * predictions: dictionary with keys being the phases provided and values are
                list the size of the number of folds in cv and containing the output of
                the estimator on that corresponding fold.
                For example, predictions["valid"][3] contains the prediction of the estimator on
                untrained data of the third fold (i.e. validation set of the third fold)
            * models: list the size of the number of folds in cv containing the fitted estimator
                on the training data of that fold.
    """
    predictor = BasePredictor.from_estimator(estimator)(estimator)
    predictions = {phase: [] for phase in phases}
    models = []
    for train_idx, valid_idx in cv:
        data = {
            "train": {
                "X": X.iloc[train_idx],
                "a": a.iloc[train_idx],
                "y": y.iloc[train_idx],
            },
            "valid": {
                "X": X.iloc[valid_idx],
                "a": a.iloc[valid_idx],
                "y": y.iloc[valid_idx],
            },
        }
        # TODO: use dict-comprehension to map between phases[0] to cv[0]
        # instead of writing "train" explicitly

        if refit:
            predictor.fit(
                X=data["train"]["X"], a=data["train"]["a"], y=data["train"]["y"]
            )

        for phase in phases:
            fold_prediction = predictor.predict(X=data[phase]["X"], a=data[phase]["a"])
            predictions[phase].append(fold_prediction)

        models.append(deepcopy(predictor.estimator))
    return predictions, models


class BasePredictor:
    """Generate predictions from estimator for evaluation (base class)."""

    @staticmethod
    def from_estimator(
        estimator: Union[
            IndividualOutcomeEstimator, PropensityEstimator, WeightEstimator
        ]
    ):
        """Select subclass based on estimator.

        Args:
            estimator (Union[IndividualOutcomeEstimator, PropensityEstimator, WeightEstimator]):
                Estimator to generate evaluation predictions from.

        Returns:
            Union[PropensityPredictor, WeightPredictor, OutcomePredictor]: the correct predictor for
                the supplied estimator
        """
        # import outside toplevel is the price you pay for having a factory method
        # of the base class

        if isinstance(estimator, PropensityEstimator):
            return PropensityPredictor
        if isinstance(estimator, WeightEstimator):
            return WeightPredictor
        if isinstance(estimator, IndividualOutcomeEstimator):
            return OutcomePredictor
        raise ValueError(f"Received unsupported estimator type {type(estimator)}")

    def __init__(self, estimator):
        self.estimator = estimator

    @abc.abstractmethod
    def fit(self, X, a, y):
        """Fit an estimator."""
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self, X, a):
        """Predict (weights, outcomes, etc. depending on the model).
        The output can be as flexible as desired, but score_estimation should know to handle it."""
        raise NotImplementedError


class OutcomePredictor(BasePredictor):
    """Generate evaluation predictions for IndividualOutcomeEstimator models."""

    def __init__(self, estimator):
        """
        Args:
            estimator (IndividualOutcomeEstimator):
        """
        if not isinstance(estimator, IndividualOutcomeEstimator):
            raise TypeError(
                f"OutcomePredictor must be initialized with IndividualOutcomeEstimator. "
                f"Received ({ type(estimator)}) instead."
            )
        super().__init__(estimator)

    def fit(self, X, a, y):
        """Fit estimator."""
        self.estimator.fit(X=X, a=a, y=y)

    def predict(self, X, a):
        """Predict on data."""
        prediction = self.estimator.estimate_individual_outcome(
            X, a, predict_proba=False
        )
        # Use predict_probability if possible since it is needed for most evaluations:
        prediction_event_prob = self.estimator.estimate_individual_outcome(
            X, a, predict_proba=True
        )
        fold_prediction = OutcomePredictions(prediction, prediction_event_prob)
        return fold_prediction


class WeightPredictor(BasePredictor):
    """Generate evaluation predictions for WeightEstimator models."""

    def __init__(self, estimator):
        """
        Args:
            estimator (WeightEstimator):
        """
        if not isinstance(estimator, WeightEstimator):
            raise TypeError(
                "WeightPredictor must be initialized with WeightEstimator."
                f"Received got ({type(estimator)}) instead."
            )
        super().__init__(estimator)

    def fit(self, X, a, y=None):
        """Fit estimator. `y` is ignored."""
        self.estimator.fit(X=X, a=a)

    def predict(self, X, a):
        """Predict on data.

        Args:
            X (pd.DataFrame): Covariates.
            a (pd.Series): Target variable - treatment assignment

        Returns:
            WeightEvaluatorPredictions
        """
        weight_by_treatment_assignment = self.estimator.compute_weights(
            X, a, treatment_values=None, use_stabilized=False
        )
        weight_for_being_treated = self.estimator.compute_weights(
            X, a, treatment_values=a.max(), use_stabilized=False
        )
        treatment_assignment_pred = self.estimator.learner.predict(
            X
        )  # TODO: maybe add predict_label to interface instead
        treatment_assignment_pred = pd.Series(treatment_assignment_pred, index=X.index)

        prediction = WeightPredictions(
            weight_by_treatment_assignment,
            weight_for_being_treated,
            treatment_assignment_pred,
        )
        return prediction


class PropensityPredictor(WeightPredictor):
    """Generate evaluation predictions for PropensityEstimator models."""

    def __init__(self, estimator):
        """
        Args:
            estimator (PropensityEstimator):
        """
        if not isinstance(estimator, PropensityEstimator):
            raise TypeError(
                "PropensityPredictor must be initialized with PropensityEstimator. "
                f"Received ({type(estimator)}) instead."
            )
        super().__init__(estimator)

    def predict(self, X, a):
        """Predict on data.

        Args:
            X (pd.DataFrame): Covariates.
            a (pd.Series): Target variable - treatment assignment

        Returns:
            PropensityEvaluatorPredictions
        """
        propensity = self.estimator.compute_propensity(X, a, treatment_values=a.max())
        propensity_matrix = self.estimator.compute_propensity_matrix(X)
        propensity_by_treatment_assignment = robust_lookup(propensity_matrix, a)

        weight_prediction = super().predict(X, a)
        # Do not force stabilize=False as in WeightEvaluator:
        weight_by_treatment_assignment = self.estimator.compute_weights(X, a)
        prediction = PropensityPredictions(
            weight_by_treatment_assignment,
            weight_prediction.weight_for_being_treated,
            weight_prediction.treatment_assignment_pred,
            propensity,
            propensity_by_treatment_assignment,
        )
        return prediction
